package clientCommunicator.rmi;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;

import org.json.JSONException;
import org.json.JSONObject;
import org.json.XML;

import clientCommunicator.ClientCommunicatorInterface;

import core.Core;
import core.Job;
import core.ResponseTemplateTypes;
import core.errorHelper.ErrorObject;
import core.jobManagerExceptions.InternalErrorException;
import core.jobManagerExceptions.JobManagerNotRunningException;
import core.jobManagerExceptions.NoJobExecutorException;
import core.jobManagerExceptions.QueueIsFullException;
import core.userData.UserData;
import core.userData.exceptions.DatabaseConnectionException;
import core.userData.exceptions.TokenNotFoundException;

public class RMICommunicator extends UnicastRemoteObject implements RMIServicesInterface, ClientCommunicatorInterface{

	private static final long serialVersionUID = 1;
	private Object waiter;
	
	public RMICommunicator() throws RemoteException{
		Core.getCore().getPluginLogger().finest("RMICommunicator: New instance of RMICommunicator!");
		waiter = new Object();
	}

	/**
	 * This method ALWAYS returns the job. If an error occurred the result will be JSON formatted (i.e. job.statusCode != Job.STATE_COMPLETE)
	 * if not it will be the requested format.
	 * 
	 * @param json
	 * @return The finished job
	 */
	private Job executeRequest(JSONObject json){
		Job job = null;
		try {
			JSONObject opspec = json.getJSONObject("request").getJSONObject("opspec");
			if ( opspec.has("format") ){
				job = new Job(this,
					opspec.getString("token"),
					opspec.getString("op"),
					json,
					opspec.getString("format"));
			} else {
				job = new Job(this,
						opspec.getString("token"),
						opspec.getString("op"),
						json);
			}
		} catch ( Exception e ){
			if ( job == null ){
				job = new Job(this, "no", "no", json); 
			}
			Core.getCore().getPluginLogger().warning("RMICommunciator: Invalid request. Exception occured!");
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
			job.result = Core.getCore().getResponseTemplate(ResponseTemplateTypes.INVALID_REQUEST).toString();
			return job;
		}
		
		try {
			Core.getCore().queueJob(job);
		} catch (JobManagerNotRunningException e2) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, job, "Missing component (JobManager) in core.");
			job.state = Job.STATE_FAILED;
			Core.getCore().translateJob(job);
			Core.getCore().getPluginLogger().warning("RMICommunciator: JobManager not running exception.");
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e2));
			return job;
		} catch (QueueIsFullException e2) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_BUSY, job, "Server is too busy to process request.");
			job.state = Job.STATE_FAILED;
			Core.getCore().translateJob(job);Core.getCore().getPluginLogger().warning("RMICommunciator: QueueIsFull Exception");
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e2));
			return job;
		} catch (NoJobExecutorException e2) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_NO_JOBEXECUTOR, job, "Server cannot process the request.");
			job.state = Job.STATE_FAILED;
			Core.getCore().translateJob(job);
			Core.getCore().translateJob(job);Core.getCore().getPluginLogger().warning("RMICommunciator: No JobExecutor for "+job);
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e2));
			return job;
		} catch (InternalErrorException e2) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, job, "Internal server error.");
			job.state = Job.STATE_FAILED;
			Core.getCore().translateJob(job);
			Core.getCore().translateJob(job);Core.getCore().getPluginLogger().warning("RMICommunciator: Internal error");
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e2));
			return job;
		}
		
		try {
			synchronized(waiter){
				waiter.wait();
			}
			Core.getCore().getPluginLogger().fine("RMICommunciator: Returning result to client!");
		} catch (InterruptedException e) {
			//Interrupted. Send error to client
			Core.getCore().getPluginLogger().warning("RMICommunciator: Interrupted. Not properly notified");
			job.result = "";
		}
		
		return job;
	}
	
	@Override
	public String xmlRequest(String xmlDoc, String token) throws RemoteException {
		//Check if there are available connection slots
		if ( !Core.getCore().newConnection() ){
			Core.getCore().getPluginLogger().warning("RMICommunicator: Connection limit reached!");
			Job errorJob = new Job(this, "", "", null);
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_BUSY, errorJob, "Connection limit exceeded");
			return errorJob.result;
		}
		
		try {
			RMICommunicatorManager.coreInt.getPluginLogger().fine("RMICommunicator: New XML request received");
			RMICommunicatorManager.coreInt.getPluginLogger().finest("RMICommunicator: "+xmlDoc);
			
			JSONObject jso = null;
			try {
				 jso = XML.toJSONObject(xmlDoc);
			} catch (JSONException e) {
				Core.getCore().getPluginLogger().warning("RMICommunicator: Request format is invalid XML");
				Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
				return "Invalid xml";
			}
			
			Job res = this.executeRequest(jso);
			Core.getCore().connectionFinished();
			Core.getCore().getPluginLogger().finer("RMICommunicator: XML result: "+res.result);
			
			return res.result;
		} catch ( Exception e ){
			Core.getCore().getPluginLogger().warning("RMICommunicator: Unknown exception occured!");
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
			return "SERVER EXCEPTION";
		}
	}
	
	@Override
	public String jsonRequest(String jsonData, String token) throws RemoteException {
		//Check if there are avilable connection slots
		if ( !Core.getCore().newConnection() ){
			Core.getCore().getPluginLogger().warning("RMICommunicator: Connection limit reached!");
			Job errorJob = new Job(this, "", "", null);
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_BUSY, errorJob, "Connection limit exceeded");
			return errorJob.result;
		}
				
		try {
			RMICommunicatorManager.coreInt.getPluginLogger().fine("RMICommunicator: New JSON request recieved.");
			RMICommunicatorManager.coreInt.getPluginLogger().finest("RMICommunicator: "+jsonData);
			
			JSONObject jso = new JSONObject(jsonData);
			
			Job j = this.executeRequest(jso); 
			Core.getCore().connectionFinished();
			Core.getCore().getPluginLogger().finer("RMICommunicator: JSON result: "+j.result);
			
			return j.result;
		} catch ( Exception e ){
			Core.getCore().getPluginLogger().warning("RMICommunicator: Unknown exception occured!");
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
			return "SERVER EXCEPTION";
		}
	}

	@Override
	public boolean verifyToken(String token) throws RemoteException {
		try {
			UserData.constructFromToken(token);
		} catch (DatabaseConnectionException e) {
			return false;
		} catch (TokenNotFoundException e) {
			return false;
		}
		
		return true;
	}
	
	@Override
	public int sendAdminCommand(String cmd, HashMap<String, String> options, String token)
			throws RemoteException {
		RMICommunicatorManager.coreInt.getPluginLogger().info("RMICommunicator: New ADMIN command request received.");
		RMICommunicatorManager.coreInt.getPluginLogger().info("RMICommunicator: Command is: "+cmd);
		
		return Core.getCore().doAdminCmd(cmd, options, token);
	}

	@Override
	public String[] getAdminCommands(String token) throws RemoteException {
		RMICommunicatorManager.coreInt.getPluginLogger().info("RMICommunicator: Admin command list requested");
		
		String[] res =  Core.getCore().requestAdminCmds(token);
		if ( res == null ){
			return new String[0];
		}
		
		return res;
	}
	

	@Override
	public String ping(String msg) {
		RMICommunicatorManager.coreInt.getPluginLogger().finer("RMICommunicator: Ping message received: "+msg);
		return msg;
	}


	@Override
	public void notifyMe() {
		synchronized(waiter){
			Core.getCore().getPluginLogger().finer("RMICommunicator: I have been notified that the job is completed.");
			waiter.notifyAll();
		}
	}
}