package config;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * ConfNode defines a set of methods for reading XML configuration files and accessing them using a simple dot notation.
 * 
 * The configuration files XML structure must be valid in order to be parsed properly (no unclosed tags or similar).
 * 
 *  Example of use:
 *  Given the configuration file (pre tags appended for readability when converted to html and are NOT part of the config file):
 *  <pre>
 *  <conf>
 *  	<group>
 *  		<value>hi there</value>
 *  		<anotherValue>hi again</anotherValue>
 *  	</group>
 *  </conf>
 *  </pre>
 * 
 * @author Thomas Haugland Rudfoss
 *
 */
public class ConfNode {
	
	/**
	 * Creates a new ConfNode set from an XML file name.
	 * 
	 * @param xmlFileName The path and filename of the xml file
	 * @return
	 * @throws SAXException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 */
	public static ConfNode makeFromXML(String xmlFileName) throws SAXException, IOException, ParserConfigurationException{
		File file = new File(xmlFileName);
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = dbf.newDocumentBuilder();
		Document doc = db.parse(file);
		doc.getDocumentElement().normalize();
		
		ConfNode root = new ConfNode(doc.getDocumentElement().getTagName());
		root.buildFromNode(doc.getDocumentElement());
		return root;
	}
	
	private ArrayList<ConfNode> children;
	private ConfNode parent;
	private String value;
	private String name;

	/**
	 * Create a new ConfNode with the given name.
	 * 
	 * @param name
	 */
	private ConfNode(String name){
		this.parent = null;
		this.name = name.toLowerCase();
		this.children = new ArrayList<ConfNode>();
	}
	
	/**
	 * Create a new ConfNode with the given name and a specific ConfNode parent.
	 * 
	 * @param parent
	 * @param name
	 */
	private ConfNode(ConfNode parent, String name){
		this.parent = parent;
		this.name = name.toLowerCase();
		this.children = new ArrayList<ConfNode>();
	}
	
	/**
	 * Returns the string value of this node.
	 * @return String value of this node.
	 */
	public String val(){
		return this.value;
	}
	
	/**
	 * Returns the string value for this node (alias for val())
	 * @return String value of this node.
	 */
	public String value(){
		return this.val();
	}
	
	/**
	 * Returns the name of this configuration node.
	 * 
	 * @return
	 */
	public String name(){
		return this.name;
	}
	
	/**
	 * Checks if this ConfNode has any children.
	 * 
	 * @return
	 */
	public boolean hasChildren(){
		if ( this.children.size() > 0 ) return true;
		return false;
	}
	
	/**
	 * Returns a list of all the children of this ConfNode or an empty list if it does not have any.
	 * 
	 * @return
	 */
	public ArrayList<ConfNode> children(){
		return this.children;
	}
	
	public ConfNode parent(){
		return this.parent;
	}
	
	/**
	 * Returns true if this node is the root node (it has no parent).
	 * 
	 * @return
	 */
	public boolean isRoot(){
		if ( this.parent == null ) return true;
		return false;
	}
	
	/**
	 * Returns the first ConfNode element with a name that matches the given string.
	 * This method is recursive. But only the first child is explored.
	 * 
	 * A blank string will return this element.
	 * 
	 * @param element The name of the element to return (can be absolute: parent.child)
	 * @throws NoChildrenException When the node has no children to search
	 * @throws NoPropWithThisNameException When no property matches the name
	 * @return
	 */
	public ConfNode get(String element){
		if ( this.children.size() == 0 ) throw new NoChildrenException();
		
		if ( element.length() == 0 ) return this;
		
		String[] splitElem = element.split("\\.");
		if ( splitElem.length == 1 ){
			//Look for first child with this name
			for ( int i=0; i<this.children.size(); i++ ){
				if ( this.children.get(i).name().equalsIgnoreCase(splitElem[0]) )
					return this.children.get(i);
			}
			
			throw new NoPropWithThisNameException(splitElem[0]);
		} else {
			String childGetter = element.substring(element.indexOf(".")+1); 
			for ( int i=0; i<this.children.size(); i++ ){
				if ( this.children.get(i).name().equalsIgnoreCase(splitElem[0]) )
					return this.children.get(i).get(childGetter);
			}
			
			throw new NoPropWithThisNameException(splitElem[0]);
		}
	}
	
	/**
	 * Find the first occurence of the given name from this node and deeper (descendants).
	 * This method only works for single names. Full paths like: "parent.child" will not work (use get())
	 * 
	 * @param name
	 * @return
	 */
	public ConfNode find(String name){
		for ( int i=0; i<this.children.size(); i++ ){
			ConfNode child = this.children.get(i);
			if ( child.name().equalsIgnoreCase(name) ) return child;
			if ( child.hasChildren() ){
				ConfNode subChild = child.find(name);
				if ( subChild != null ) return subChild;
			}
		}
		
		return null;
	}
	
	/**
	 * Returns a list of all nodes that match the last element in the selector.
	 * Example:
	 * <pre>
	 * <conf>
	 * 	<job>hi</job>
	 * 	<job>there</job>
	 * 	<job>world</job>
	 * </conf>
	 * </pre>
	 * If this is the "conf" node and you call getAll("job") you will get all the job nodes
	 * <pre>
	 * <conf>
	 * 	<job>
	 * 		<op>anoption</op>
	 * 		<op>anoption</op>
	 * 	</job>
	 * 	<job>
	 * 		<op>op</op>
	 * 		<op>op</op>
	 * 	</job>
	 * 	<job>world</job>
	 * </conf>
	 * </pre>
	 * If this is the "conf" node and you call getAll("job.op") you will get all the op nodes from the FIRST JOB ONLY.
	 * 
	 * @param element
	 * @return
	 */
	public ArrayList<ConfNode> getAll(String element){
		if ( this.children.size() == 0 ) throw new NoChildrenException();
		
		ArrayList<ConfNode> result = new ArrayList<ConfNode>();
		if ( element.length() == 0 ) {
			result.add(this);
			return result;
		}
		
		String[] splitElem = element.split("\\.");
		if ( splitElem.length == 1 ){
			//Look for first child with this name
			for ( int i=0; i<this.children.size(); i++ ){
				if ( this.children.get(i).name().equalsIgnoreCase(splitElem[0]) )
					result.add(this.children.get(i));
			}
			
			if ( result.size() == 0 ) throw new NoPropWithThisNameException(splitElem[0]);
			return result;
		} else {
			String childGetter = element.substring(element.indexOf(".")+1); 
			for ( int i=0; i<this.children.size(); i++ ){
				if ( this.children.get(i).name().equalsIgnoreCase(splitElem[0]) )
					return this.children.get(i).getAll(childGetter);
			}
			
			throw new NoPropWithThisNameException(splitElem[0]);
		}
	}
	
	/**
	 * Attempts to parse the value as a boolean using several different methods.
	 * The following evaluates to true (case-insensitive):
	 * "yes"
	 * "true"
	 * "1"
	 * 
	 * Anything else evaluates to false.
	 * 
	 * @return
	 */
	public boolean asBoolean(){
		if ( this.value.equalsIgnoreCase("yes") ) return true;
		if ( this.value.equalsIgnoreCase("true") ) return true;
		if ( this.value.equalsIgnoreCase("1") ) return true;
		
		return false;
	}
	
	/**
	 * Attempts to parse the value as an Integer object. Returns null if not successful.
	 * 
	 * @return
	 */
	public Integer asInt(){
		try {
			return Integer.parseInt(this.value);
		} catch ( NumberFormatException e ){
			return null;
		}
	}
	
	/**
	 * Attempts to parse the value as a Double object. Returns null if not successful.
	 * 
	 * @return
	 */
	public Double asDouble(){
		try {
			return Double.parseDouble(this.value);
		} catch ( NumberFormatException e ){
			return null;
		}
	}
	
	/**
	 * Attempts to parse the value as a Long object. Returns null if not successful.
	 * 
	 * @return
	 */
	public Long asLong(){
		try {
			return Long.parseLong(this.value);
		} catch ( NumberFormatException e ){
			return null;
		}
	}
	
	/**
	 * Attempts to split the string into an array using the provided regular expression. Returns the result of String.split().
	 * 
	 * @return
	 */
	public String[] asArray(String regex){
		return this.value.split(regex);
	}

	/**
	 * Populates this ConfNode with data from a node in an XML document.
	 * Used during loading to create the node tree.
	 * 
	 * @param rootNode
	 */
	protected void buildFromNode(Element rootNode){
		if ( !rootNode.hasChildNodes() ) return;
		
		ArrayList<Element> validChildren = getValidChildren(rootNode); 
		if ( validChildren.size() == 0 ) {
			//This is a leaf node with only text
			this.value = rootNode.getTextContent().trim();
			return;
		}
		
		for ( int i=0; i<validChildren.size(); i++ ){
			Element childNode = (Element) validChildren.get(i);
			ConfNode newChild = new ConfNode(this, childNode.getTagName());
			newChild.buildFromNode(childNode);
			this.children.add(newChild);
		}
	}
	
	private ArrayList<Element> getValidChildren(Node node){		
		ArrayList<Element> validChildren = new ArrayList<Element>();
		int childLength = node.getChildNodes().getLength();
		for ( int u=0; u<childLength; u++ ){
			Node childNode = node.getChildNodes().item(u);
			if ( childNode == null ) continue;
			if ( childNode.getNodeType() != Node.ELEMENT_NODE ) continue;
			
			validChildren.add((Element) childNode);
		}
		
		return validChildren;
	}
	
	/**
	 * Returns a textual representation of this and all descendant nodes.
	 */
	public String toString(){
		String res = null;
		if ( this.hasChildren() ){
			res = "<"+this.name+">";
			for ( int i=0; i<this.children.size(); i++ ){
				res += "\r\n"+this.children.get(i).toString("  ");
			}
			res += "\r\n</"+this.name+">";
		} else {
			res = this.name+" = "+this.value;
		}
		
		return res;
	}
	
	protected String toString(String indent){
		String res = null;
		if ( this.hasChildren() ){
			res = indent+"<"+this.name+">";
			for ( int i=0; i<this.children.size(); i++ ){
				res += "\r\n"+indent+""+this.children.get(i).toString(indent+"  ");
			}
			res += "\r\n"+indent+"</"+this.name+">";
		} else {
			res = indent+""+this.name+" = "+this.value;
		}
		
		return res;
	}
}
