package core;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.ParserConfigurationException;

import org.json.JSONException;
import org.json.JSONObject;
import org.xml.sax.SAXException;


import clientCommunicator.ClientCommunicatorManager;

import config.ConfNode;
import config.NoPropWithThisNameException;
import core.jobManagerExceptions.InternalErrorException;
import core.jobManagerExceptions.JobManagerNotRunningException;
import core.jobManagerExceptions.NoJobExecutorException;
import core.jobManagerExceptions.QueueIsFullException;
import core.userData.UserData;
import core.userData.exceptions.DatabaseConnectionException;
import core.userData.exceptions.TokenNotFoundException;

/**
 * 
 * 
 * 
 * Admin command return codes:
 * -1 : Undefined error
 * 0 : Command completed successfully
 * 
 * 10 : Access denied (User access denied, invalid token)
 * 20 : Database error during creation of user object
 * 30 : Token not found (User doesn't exist)
 * 
 * @author Thomas Rudfoss
 *
 */
public final class Core implements CoreInterface {
	
	private static Core instance;
	
	/**
	 * Returns the active instance of the core.
	 * This can only be used by classes inside the core package.
	 * The returned Core har more methods than using getCore.
	 * 
	 * getCore is meant for use by external plugins. 
	 * 
	 * @return
	 */
	protected static Core getCoreFromCore(){
		if ( Core.instance == null ) Core.instance = new Core();
		
		return Core.instance;
	}
	
	/**
	 * Returns the core as defined by the CoreInterface.
	 * 
	 * @return The Core
	 */
	public static CoreInterface getCore(){
		if ( Core.instance == null ) Core.instance = new Core();
		
		return (CoreInterface) Core.instance;
	}
	
	
	
	private static final String appName = "Core";
	private static final double version = 0.2;
	
	private static final String configFilePath = "config.xml";
	
	
	
	public static final int STATE_INITIALIZING = 1;
	public static final int STATE_INITFAILURE = 3;
	public static final int STATE_RUNNING = 5;
	public static final int STATE_TERMINATING = 10;
	
	/**
	 * The state of the core. This is not accessible from outside. Use getState.
	 * See Core.STATE_* for states the core can have.
	 */
	private static int state = 0;
	
	/**
	 * Contains all config parameters for the system.
	 * This object is immediatly available to all classes inside the Core package because we can trust these.
	 */
	protected static ConfNode config;
	protected static JobManager jobManager;
	private static Object blankObject;
	
	/**
	 * Holds a reference to the jobExecutor class that can perform a given job.
	 * 
	 * getPerson -> getPersonClass (reference)
	 */
	protected static HashMap<String, Class<?>> jobExecutorList;
	
	private static ArrayList<ClientCommunicatorManager> clientCommManagers;
	
	protected static Logger logger;
	protected static Logger pluginLogger;
	
	
	
	
	
	
	
	private int connectionCount;
	private int maxConnections;

	//Core methods
	private Core(){}
	
	/**
	 * Initialize the core, load plugins and ready job manager.
	 * Basically: Call this FIRST to start the system.
	 */
	public synchronized void begin(){
		if ( Core.state != 0 ) return;
		Core.state = Core.STATE_INITIALIZING;
		
		System.out.println(Core.appName+" v"+Core.version+"");
		System.out.println("");
		
		this.init();
		if ( Core.state == Core.STATE_INITIALIZING ) this.loadPlugins();
		if ( Core.state == Core.STATE_INITIALIZING ) this.run();
		if ( Core.state != Core.STATE_INITIALIZING ){
			Core.logger.warning("Core: Some error occured during initialization. Cannot start.");
		}
		this.terminate();
	}
	
	private void init(){
		//Load config file
		try {
			Core.config = ConfNode.makeFromXML(Core.configFilePath);
			Core.blankObject = new Object();
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
			Core.state = Core.STATE_INITFAILURE;
		} catch (SAXException e) {
			e.printStackTrace();
			Core.state = Core.STATE_INITFAILURE;
		} catch (IOException e) {
			e.printStackTrace();
			Core.state = Core.STATE_INITFAILURE;
		}
		
		if ( Core.state == Core.STATE_INITFAILURE){
			return;
		}
		
		this.init_setupLogging();
		
		Core.jobManager = new JobManager();
		if ( !Core.jobManager.active ) {
			Core.logger.warning("Core: Could not start job manager.");
			Core.state = Core.STATE_INITFAILURE;
		}
		
		Core.jobExecutorList = new HashMap<String, Class<?>>();
		Core.clientCommManagers = new ArrayList<ClientCommunicatorManager>();
		
		this.connectionCount = 0;
		this.maxConnections = Core.config.get("core.maxIncommingConnections").asInt();
		
		UserData.initUserData(jobManager.coreDBCon);
	}
	
	private void init_setupLogging(){
		//Initialize the core logger
		Core.logger = Logger.getLogger("core");
		boolean coreLogging = Core.config.get("logging.enableCoreLogger").asBoolean();
		if ( !coreLogging ) {
			Core.logger.setLevel(Level.OFF);
		} else {
			String logLevel = Core.config.get("logging.coreLogLevel").value();
			
			if ( logLevel.equalsIgnoreCase("finest") ) Core.logger.setLevel(Level.FINEST);
			else if ( logLevel.equalsIgnoreCase("finer") ) Core.logger.setLevel(Level.FINER);
			else if ( logLevel.equalsIgnoreCase("fine") ) Core.logger.setLevel(Level.FINE);
			else if ( logLevel.equalsIgnoreCase("config") ) Core.logger.setLevel(Level.CONFIG);
			else if ( logLevel.equalsIgnoreCase("info") ) Core.logger.setLevel(Level.INFO);
			else if ( logLevel.equalsIgnoreCase("warning") ) Core.logger.setLevel(Level.WARNING);
			else if ( logLevel.equalsIgnoreCase("severe") ) Core.logger.setLevel(Level.SEVERE);
			else if ( logLevel.equalsIgnoreCase("off") ) Core.logger.setLevel(Level.OFF);
			
			Handler[] handlers = Logger.getLogger("").getHandlers();
			for ( int i=0; i<handlers.length; i++ ){
				handlers[i].setLevel(Core.logger.getLevel());
			}
			
			Core.logger.info("Core: Logging active (Level: "+logLevel+")");
		}
		
		//Initialize the plugin logger
		Core.pluginLogger = Logger.getLogger("Core.plugin");
		boolean pluginLoggging = Core.config.get("logging.enablePluginsLogger").asBoolean();
		if ( !pluginLoggging ) {
			Core.pluginLogger.setLevel(Level.OFF);
			Core.logger.info("Core: Plugin logging disabled");
		} else {
			String logLevel = Core.config.get("logging.pluginsLogLevel").value();
			
			if ( logLevel.equalsIgnoreCase("inherit") ) logLevel = Core.config.get("logging.coreLogLevel").value();
			
			if ( logLevel.equalsIgnoreCase("finest") ) Core.pluginLogger.setLevel(Level.FINEST);
			else if ( logLevel.equalsIgnoreCase("finer") ) Core.pluginLogger.setLevel(Level.FINER);
			else if ( logLevel.equalsIgnoreCase("fine") ) Core.pluginLogger.setLevel(Level.FINE);
			else if ( logLevel.equalsIgnoreCase("config") ) Core.pluginLogger.setLevel(Level.CONFIG);
			else if ( logLevel.equalsIgnoreCase("info") ) Core.pluginLogger.setLevel(Level.INFO);
			else if ( logLevel.equalsIgnoreCase("warning") ) Core.pluginLogger.setLevel(Level.WARNING);
			else if ( logLevel.equalsIgnoreCase("severe") ) Core.pluginLogger.setLevel(Level.SEVERE);
			else if ( logLevel.equalsIgnoreCase("off") ) Core.pluginLogger.setLevel(Level.OFF);
			
			Handler[] handlers = Core.pluginLogger.getHandlers();
			for ( int i=0; i<handlers.length; i++ ){
				handlers[i].setLevel(Core.pluginLogger.getLevel());
			}
			
			Core.logger.info("Core: Plugin logging active (Level: "+logLevel+")");
		}
	}

	/**
	 * Loads all plugins defined in the config file.
	 * If the plugin entry does not have a jarPath entry it is assumed to exist internally and new instance
	 * is called on the className class.
	 */
	private void loadPlugins(){
		Core.logger.info("Core: Loading plugins...");
		
		ConfNode rootPlugin = Core.config.get("plugins");
		
		//Get all plugins that have an external jar to simplify loading them later.
		ConfNode clientCommunicators = rootPlugin.get("clientCommunicators");
		ConfNode jobExecutorPlugins = rootPlugin.get("jobExecutors");
		ConfNode translatorPlugins = rootPlugin.get("resultTranslators");
		
		ArrayList<URL> pluginJarPaths = new ArrayList<URL>(); //Holds the paths to all external jar files that are to be loaded
		if ( clientCommunicators.hasChildren() ){ //Build a list of external jar paths for easy loading later.
			for ( int i=0; i<clientCommunicators.children().size(); i++ ){
				String path = null;
				try {
					path = clientCommunicators.children().get(i).get("jarPath").val();
					pluginJarPaths.add((new File(path)).toURI().toURL());
				} catch ( NoPropWithThisNameException e ){} catch (MalformedURLException e) {
					Core.logger.warning("Core: Could not load "+path+". Malformed URL exception!");
				}
			}
		}
		
		if ( jobExecutorPlugins.hasChildren() ){ //Build a list of external jar paths for easy loading later.
			for ( int i=0; i<jobExecutorPlugins.children().size(); i++ ){
				String path = null;
				try {
					path = jobExecutorPlugins.children().get(i).get("jarPath").val();
					pluginJarPaths.add((new File(path)).toURI().toURL());
				} catch ( NoPropWithThisNameException e ){} catch (MalformedURLException e) {
					Core.logger.warning("Core: Could not load "+path+". Malformed URL exception!");
				}
			}
		}
		
		if ( translatorPlugins.hasChildren() ){ //Build a list of external jar paths for easy loading later.
			for ( int i=0; i<translatorPlugins.children().size(); i++ ){
				String path = null;
				try {
					path = translatorPlugins.children().get(i).get("jarPath").val();
					pluginJarPaths.add((new File(path)).toURI().toURL());
				} catch ( NoPropWithThisNameException e ){} catch (MalformedURLException e) {
					Core.logger.warning("Core: Could not load "+path+". Malformed URL exception!");
				}
			}
		}
		
		URL[] jarPathsAsURLs = new URL[pluginJarPaths.size()];
		jarPathsAsURLs = pluginJarPaths.toArray(jarPathsAsURLs);
		
		//Make a URLClassLoader for all external jar files
		URLClassLoader externalJarLoader = new URLClassLoader(jarPathsAsURLs, this.getClass().getClassLoader());
		
		
		
		
		
		//Ready to begin loading plugins
		//Loading jobExecutor plugins
		if ( jobExecutorPlugins.hasChildren() ){
			Core.logger.info("Core: Loading JobExecutors ("+jobExecutorPlugins.children().size()+")...");
			for ( int i=0; i<jobExecutorPlugins.children().size(); i++ ){
				ConfNode thisChild = jobExecutorPlugins.children().get(i);
				Class<?> plugin = null;
				try {
					try {
						thisChild.get("jarPath");
						//If it gets this far then the class is located in external jar
						plugin = Class.forName(thisChild.get("classPath").val(), true, externalJarLoader);
					} catch ( NoPropWithThisNameException e ){
						//Class is already in the classPath. No need to load exeternal jar.
						plugin = Class.forName(thisChild.get("classPath").val());
					}
					
					String[] jobs = (String[]) plugin.getMethod("jobsICanDo").invoke((Object)null);
					
					for ( int u=0; u<jobs.length; u++ ){
						String thisJob = jobs[u];
						Core.jobExecutorList.put(thisJob.toLowerCase(), plugin);
						Core.logger.fine("Core: Added jobExecutor for "+thisJob);
					}
					
					Core.logger.info("Core: Added jobExecutor "+thisChild.get("classPath").val());
				} catch (ClassNotFoundException e1) {
					Core.logger.warning("Core: Could not load "+thisChild.get("classPath").val()+" Class not found!");
					continue;
				} catch (IllegalAccessException e1) {
					Core.logger.warning("IllegalAccessException occured!");
					Core.logger.warning(Core.stackTraceToString(e1));
					continue;
				} catch (ClassCastException e1){
					Core.logger.warning("IllegalAccessException occured!");
					Core.logger.warning(Core.stackTraceToString(e1));
					continue;
				} catch (IllegalArgumentException e) {
					Core.logger.warning("IllegalArgumentException occured!");
					Core.logger.warning(Core.stackTraceToString(e));
				} catch (SecurityException e) {
					Core.logger.warning("SecurityException occured!");
					Core.logger.warning(Core.stackTraceToString(e));
				} catch (InvocationTargetException e) {
					Core.logger.warning("InvocationTargetException occured!");
					Core.logger.warning(Core.stackTraceToString(e));
				} catch (NoSuchMethodException e) {
					Core.logger.warning("NoSuchMethodException occured!");
					Core.logger.warning(Core.stackTraceToString(e));
				}
			}
		}
		
		//Load resultTranslator plugins
		if ( translatorPlugins.hasChildren() ){
			Core.logger.info("Core: Loading ResultTranslators ("+translatorPlugins.children().size()+")...");
			for ( int i=0; i<translatorPlugins.children().size(); i++ ){
				ConfNode thisChild = translatorPlugins.children().get(i);
				Class<?> plugin = null;
				try {
					try {
						thisChild.get("jarPath");
						//If it gets this far then the class is located in external jar
						plugin = Class.forName(thisChild.get("classPath").val(), true, externalJarLoader);
					} catch ( NoPropWithThisNameException e ){
						//Class is already in the classPath. No need to load exeternal jar.
						plugin = Class.forName(thisChild.get("classPath").val());
					}
					
					if ( !ResultTranslatorWrapper.addResultTranslator(plugin) ){
						Core.logger.warning("Core: Could not load "+thisChild.get("classPath").val());
					} else {
						Core.logger.info("Core: Added resultTranslator "+thisChild.get("classPath").val());
					}
				} catch (ClassNotFoundException e1) {
					Core.logger.warning("Core: Could not load "+thisChild.get("classPath").val()+" Class not found!");
					continue;
				} catch (ClassCastException e1){
					Core.logger.warning("IllegalAccessException occured!");
					Core.logger.warning(Core.stackTraceToString(e1));
					continue;
				} catch (IllegalArgumentException e) {
					Core.logger.warning("IllegalArgumentException occured!");
					Core.logger.warning(Core.stackTraceToString(e));
				} catch (SecurityException e) {
					Core.logger.warning("SecurityException occured!");
					Core.logger.warning(Core.stackTraceToString(e));
				}
			}
		}
		
		//Load clientCommunicator plugins
		if ( clientCommunicators.hasChildren() ){
			Core.logger.info("Core: Loading ClientCommunicators ("+clientCommunicators.children().size()+")...");
			for ( int i=0; i<clientCommunicators.children().size(); i++ ){
				ConfNode thisChild = clientCommunicators.children().get(i);
				Class<?> plugin = null;
				ClientCommunicatorManager pluginInstance = null;
				try {
					try {
						thisChild.get("jarPath");
						//If it gets this far then the class is located in external jar
						plugin = Class.forName(thisChild.get("classPath").val(), true, externalJarLoader);
						Core.logger.info("Core: Loaded ClientCommunicator from "+thisChild.get("jarPath"));
					} catch ( NoPropWithThisNameException e ){
						//Class is already in the classPath. No need to load exeternal jar.
						plugin = Class.forName(thisChild.get("classPath").val());
					}

					pluginInstance = (ClientCommunicatorManager) plugin.newInstance();
				} catch (ClassNotFoundException e1) {
					Core.logger.warning("Core: Could not load "+thisChild.get("classPath").val()+" Class not found!");
					continue;
				} catch (InstantiationException e1) {
					Core.logger.warning("InstantiationException occured!");
					Core.logger.warning(Core.stackTraceToString(e1));
					continue;
				} catch (IllegalAccessException e1) {
					Core.logger.warning("IllegalAccessException occured!");
					Core.logger.warning(Core.stackTraceToString(e1));
					continue;
				} catch (ClassCastException e1){
					Core.logger.warning("IllegalAccessException occured!");
					Core.logger.warning(Core.stackTraceToString(e1));
					continue;
				}
				
				Core.clientCommManagers.add(pluginInstance);
				pluginInstance.start();
			}
		}
		
		
		
		Core.logger.info("Core: Plugins loaded");
	}
	
	private void run(){
		Core.logger.info("Core: Ready to receive jobs.");
		
		if ( Core.state == Core.STATE_INITIALIZING ){
			Core.state = Core.STATE_RUNNING;
			Core.logger.info("Core: Going into suspend...");
			try {
				if ( Core.config.get("core.enableConsoleControls").asBoolean() ) (new Thread(new ConsoleAdmin())).start();
				synchronized(Core.blankObject){
					Core.blankObject.wait();
				}
				Core.logger.info("Core: Awakening from suspend. Procceeding to shut down the application.");
			} catch (InterruptedException e) {Core.logger.warning("Core: Interrupted during suspend.");}
		}
	}
	
	/**
	 * External classes in the core package may use this method to terminate the core.
	 */
	protected void shutdown(){
		synchronized(Core.blankObject){
			Core.blankObject.notify();
			Core.logger.info("Core: Notification sent. Core will awake and shut down.");
		}
	}
	
	/**
	 * This method is used internally by the core during termination.
	 */
	private void terminate(){
		Core.state = Core.STATE_TERMINATING;
		
		Iterator<ClientCommunicatorManager> it = Core.clientCommManagers.iterator();
		while ( it.hasNext() ){
			it.next().stop();
		}
		
		if ( Core.state == Core.STATE_INITFAILURE ){
			Core.logger.warning("Core: Terminating with failure");
			System.err.println("Abnormal termination");
			System.exit(1);
		}
		
		Core.logger.info("Core: Normal termination");
		System.exit(0);
	}
	
	
	
	
	
	
	//Protected methods
	protected DBConnection getCoreDBConnection(){
		return jobManager.coreDBCon;
	}
	
	
	
	
	
	
	
	
	
	
	//Externally visible methods
	
	/**
	 * Returns the state of the core (See Core.STATE_[x] constants)
	 */
	public int getState(){
		return Core.state;
	}
	
	/**
	 * Queue up a new job in the system. Used by clientCommunicators to queue up new jobs.
	 */
	public void queueJob(Job job) throws JobManagerNotRunningException, QueueIsFullException, NoJobExecutorException, InternalErrorException{
		Core.jobManager.queueJob(job);
	}
	
	/**
	 * Returns the root ConfNode for all config parameters
	 */
	public ConfNode getConfig(){
		return Core.config;
	}
	
	/**
	 * Returns the root ConfNode for a given plugin by name.
	 */
	public ConfNode getConfigForPluginByName(String name){
		ConfNode plugins = config.get("plugins");
		plugins = plugins.find("name");
		if ( plugins == null ) return null;
		if ( plugins.val().equalsIgnoreCase(name) ) return plugins.parent();
		
		return null;
	}
	
	/**
	 * Returns the logger for plugins. External plugins should use this if they want to log anything.
	 */
	public Logger getPluginLogger(){
		return Core.pluginLogger;
	}

	@Override
	public void connectionFinished() {
		Core.logger.fine("Core: Connection released. "+this.connectionCount+" connections left.");
		this.connectionCount--;
		if ( this.connectionCount < 0 ){
			this.connectionCount = 0;
			Core.logger.warning("Core: Connection count became negative. Some plugin may be calling connectionFinished without first calling newConnection.");
		}
	}

	@Override
	public boolean newConnection() {
		if ( this.connectionCount < this.maxConnections ){
			this.connectionCount++;
			Core.logger.fine("Core: Connection request granted. "+this.connectionCount+" connections left.");
			return true;
		}
		
		Core.logger.warning("Core: Connection request NOT granted.");
		return false;
	}
	
	/**
	 * Performs administrative commands if the user is valid.
	 */
	public int doAdminCmd(String cmd, HashMap<String, String> options, String token){
		UserData ud = null;
		try {
			ud = UserData.constructFromToken(token);
		} catch (DatabaseConnectionException e) {
			Core.logger.warning("Core: Could not execute admin command. Database error!");
			Core.logger.warning(Core.stackTraceToString(e));
			return 20;
		} catch (TokenNotFoundException e) {
			Core.logger.warning("Core: Could not execute admin command. Token not found.");
			return 30;
		}
		
		if ( !ud.isAdmin() ){
			return 10;
		}
		
		if ( cmd.equalsIgnoreCase("shutdown") ){
			Core.logger.info("Core: Admin command "+cmd+" received. User verified. Shutting down...");
			this.shutdown();
		} //Add more commands here
		
		/*
		 * Potential commands:
		 * - GetAllTokens
		 * - GetInfo
		 * 		- What modules are loaded
		 * 		- What tasks the system can perform
		 * 		- What what translations the system can do
		 */

		return 0;
	}
	
	public String[] requestAdminCmds(String token){
		UserData ud = null;
		
		try {
			ud = UserData.constructFromToken(token);
		} catch (DatabaseConnectionException e) {
			Core.logger.warning("Core: Could not execute admin command. Database error!");
			Core.logger.warning(Core.stackTraceToString(e));
			return null;
		} catch (TokenNotFoundException e) {
			Core.logger.warning("Core: Could not execute admin command. Token not found.");
			return null;
		}
		
		if ( !ud.isAdmin() ){
			return null;
		}
		
		return new String[]{"Shutdown"};
	}
	
	@Override
	@Deprecated
	public JSONObject getResponseTemplate(ResponseTemplateTypes templateType){
		try {
			return new JSONObject(templateType.json);
		} catch (JSONException e) {
			Core.logger.severe("Core: Could not construct template from: "+templateType.json);
			Core.logger.severe(Core.stackTraceToString(e));
		}
		return null;
	}
	
	@Override
	public void translateJob(Job job){
		Core.jobManager.translateResult(job);
	}
	
	
	
	
	
	
	
	
	/**
	 * Extracts a stack trace from the given exception and returns it as a string.
	 * 
	 * @param e Exception to extract stack trace from
	 * @return String equal to what e.printStackTrace() dumps to standard out
	 */
	public static String stackTraceToString(Exception e){
		String val = "Failed to get stack trace.";
		StringWriter sw = null;
		PrintWriter pw = null;
		
		try{
			sw = new StringWriter();
			pw = new PrintWriter(sw);
			e.printStackTrace(pw);
			val = sw.toString();
		} finally {
			try {
				if ( pw != null ) pw.close();
				if ( sw != null ) sw.close();
			} catch (IOException ignore) {}
		}
		
		return val;
	}

}
