package core;

import java.util.HashMap;
import java.util.logging.Logger;

import org.json.JSONObject;

import config.ConfNode;
import core.jobManagerExceptions.InternalErrorException;
import core.jobManagerExceptions.JobManagerNotRunningException;
import core.jobManagerExceptions.NoJobExecutorException;
import core.jobManagerExceptions.QueueIsFullException;

/**
 * CoreInterface defines how external plugins should communicate with the core.
 * 
 * @author Thomas Rudfoss
 *
 */
public interface CoreInterface {

	/**
	 * Returns the state of the core (See Core.STATE_[x] constants)
	 */
	public int getState();
	
	/**
	 * Queue a job for execution.
	 * 
	 * Use this from any ClientCommunicator to queue up a new job. Once queued the ClientCommunicator should sleep.
	 * When the ClientCommunicator is notified the job is either complete (including translations) or failed for some reason.
	 * 
	 * @param job The job that will be queued.
	 * @throws JobManagerNotRunningException
	 * @throws QueueIsFullException
	 * @throws NoJobExecutorException
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	public void queueJob(Job job) throws JobManagerNotRunningException, QueueIsFullException, NoJobExecutorException, InternalErrorException;
	
	/**
	 * Returns all configuration parameters in a ConfNode object 
	 * @return
	 */
	public ConfNode getConfig();
	
	/**
	 * Returns the root ConfNode (plugin) for a plugin with the given name. Works for all types of plugins.
	 * 
	 * @param name
	 * @return
	 */
	public ConfNode getConfigForPluginByName(String name);
	
	/**
	 * Returns the logger for plugins. Use this to log any event that occurs within a plugin.
	 * 
	 * @return
	 */
	public Logger getPluginLogger();
	
	/**
	 * Checks if there are available slots for new connections.
	 * If there are available slots one will be reserved when this method is called. Therefore if the ClientCommunicator
	 * does not wish to use the connection it must immediately call connectionFinished to free up the slot. 
	 * 
	 * ClientCommunicators can use this before allowing new connections into the system.
	 * 
	 * @return
	 */
	public boolean newConnection();
	
	/**
	 * Releases a connection slot back to the pool. There is no check done on whether the connection remains open so it is once
	 * again up to the plugin to do this.
	 * 
	 * Be sure to call this or the system will fast begin to ignore connections.
	 */
	public void connectionFinished();
	
	/**
	 * Perform an administrative command on the system. Requires an admin user token be present in the AdminCommand object.
	 * 
	 * @param cmd The command object to execute
	 */
	public int doAdminCmd(String cmd, HashMap<String, String> options, String token);
	
	/**
	 * Request a list of admin commands
	 * @param token
	 * @return
	 */
	public String[] requestAdminCmds(String token);
	
	/**
	 * Returns the requested template for a response. This simplifies the construction of responses.
	 * Use constants defined in the enum ResponseTemplateTypes.
	 * 
	 * @param templateType
	 * @return
	 */
	public JSONObject getResponseTemplate(ResponseTemplateTypes templateType);
	
	/**
	 * Perform a translation of the result from outside the JobExecutor.
	 * @param job
	 */
	public void translateJob(Job job);
}
