package core;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Iterator;

public class DBConnection {

	private final Connection con;
	
	private ArrayList<Statement> statements;
	
	private boolean closing;
	
	protected DBConnection(Connection con){
		this.con = con;
		this.closing = false;
		this.statements = new ArrayList<Statement>();
	}
	
	/**
	 * Terminates the connection and closes any open statements.
	 * Statements that are open and currently being executed are canceled before they are closed to ensure consistency.
	 * 
	 * @throws SQLException
	 */
	protected synchronized void close() throws SQLException{
		this.closing = true;
		Iterator<Statement> it = this.statements.iterator();
		while ( it.hasNext() ){
			Statement thisStatement = it.next();
			
			if ( !thisStatement.isClosed() ){
				thisStatement.cancel();
				thisStatement.close();
			}
		}
		
		this.con.commit();
		this.con.close();
	}
	
	/**
	 * Clean up the database connection by closing all active statements.
	 * 
	 * @throws SQLException
	 */
	protected void clean() throws SQLException{
		this.closing = true;
		Iterator<Statement> it = this.statements.iterator();
		while ( it.hasNext() ){
			Statement thisStatement = it.next();
			
			if ( !thisStatement.isClosed() ){
				thisStatement.cancel();
				thisStatement.close();
			}
		}
		
		this.closing = false;
	}
	
	
	
	
	
	
	
	/**
	 * Returns a new statement object for the database connection. Statements that are not closed by the plugin are automatically
	 * closed when the plugin finishes its work.
	 * 
	 * Warning:
	 * If the DBConnection is in the process of closing (either close() or clean() has been called) a null object will be returned.
	 * 
	 * @return A new statement object.
	 * @throws SQLException
	 */
	public synchronized Statement newStatement() throws SQLException{
		if ( this.closing ) return null;
		return con.createStatement(ResultSet.TYPE_SCROLL_SENSITIVE, ResultSet.CONCUR_UPDATABLE, ResultSet.CLOSE_CURSORS_AT_COMMIT);
	}
	
	/**
	 * Creates a new prepared statement from the given query and returns a reference to it.
	 * 
	 * Warning:
	 * If the DBConnection is in the process of closing (either close() or clean() has been called) a null object will be returned.
	 * 
	 * @param sql An sql query to put in this statement.
	 * @return A new statement object.
	 * @throws SQLException
	 */
	public synchronized PreparedStatement newPreparedStatement(String sql) throws SQLException{
		if ( this.closing ) return null;
		
		return con.prepareStatement(sql, ResultSet.TYPE_SCROLL_SENSITIVE, ResultSet.CONCUR_UPDATABLE);
	}
}
