package core;

import org.json.JSONObject;

import core.errorHelper.ErrorObject;
import core.userData.UserData;

import resultTranslators.ResultTranslatorInterface;
import clientCommunicator.ClientCommunicatorInterface;
import jobExecutors.JobExecutorInterface;

/**
 * 
 * Job
 * 
 * @author Thomas Rudfoss
 *
 */
public class Job {
	
	/**
	 * This state defines a new job that has not begun executing
	 */
	public static final int STATE_NEW = 1;
	/**
	 * The state for a job that is executing.
	 */
	public static final int STATE_RUNNING = 10;
	/**
	 * A job that has a result set (before translated)
	 */
	public static final int STATE_RESULT_SET = 20;
	/**
	 * A job that has a translated result
	 */
	public static final int STATE_RESULT_TRANSLATED = 30;
	/**
	 * Job that is complete
	 */
	public static final int STATE_COMPLETE = 50;
	/**
	 * The job has been cancelled before it had a chance to execute
	 */
	public static final int STATE_CANCELED = 60;
	
	public static final int STATE_TIMEOUT = 100;
	public static final int STATE_FAILED = 200;
	
	private static int lastId = 0;
	private synchronized static int getNextId(){
		Job.lastId++;
		return lastId;
	}
	
	/**
	 * A unique ID for this
	 */
	public final int jobId;
	
	protected ClientCommunicatorInterface communicator;
	protected JobExecutorInterface executor;
	protected ResultTranslatorInterface[] translators;
	public DBConnection dbConnection;
	
	private String jobTask;
	private String token;
	/**
	 * A UserData object describing the user that requested the job.
	 */
	public UserData userData;
	public int state;
	/**
	 * The complete request as a JSONObject
	 */
	public JSONObject request;
	public String resultFormat;
	
	/**
	 * The internal result of the job using a format that the system can understand (probably JSON).
	 */
	public JSONObject internalResult;
	/**
	 * The final result of the job.
	 */
	public String result;
	/**
	 * An error object ONLY IF an error occurred. Else null.
	 */
	public ErrorObject error;
	
	/**
	 * Contains a status message explaining the status of the job.
	 */
	public String statusMessage;
	/**
	 * Contains a status code defining the job status.
	 */
	public int statusCode;
	
	/**
	 * A unix timestamp defining when the job was created in milliseconds.
	 */
	public long time_created;
	/**
	 * A unix timestamp defining when the job was queued for execution in milliseconds.
	 */
	public long time_queued;
	/**
	 * A unix timestamp defining when the job was dequeued and allowed to begin executing in milliseconds.
	 */
	public long time_dequeued;
	/**
	 * A unix timestamp defining when the job was started in the jobexecutor.
	 */
	public long time_jobStart;
	/**
	 * A unix timestamp defining when the job was finished in the jobexecutor.
	 */
	public long time_jobFinished;
	/**
	 * A unix timestamp defining when the job was completed in milliseconds.
	 */
	public long time_completed;
	/**
	 * A unix timestamp defining when the job was fully translated (aka. ready to return to client) in milliseconds.
	 */
	public long time_translated;

	/**
	 * Creates a new Job object with the given parameters
	 * 
	 * @param communicator The ClientCommuncator that will be associated with this job. Only this communicator will be notified when the job is complete (or fails).
	 * @param jobTask The name of the task the job should execute (e.g.: "getClassParticipants").
	 * @param request The complete request as a JSONObject
	 * @param resultFormat The requested format for the result.
	 */
	public Job(ClientCommunicatorInterface communicator, String token, String jobTask, JSONObject request, String resultFormat){
		this.jobId = getNextId();
		this.init(communicator, jobTask, token, request, resultFormat);
	}
	
	/**
	 * Works like the other constructor except resultFormat is set to JSON
	 * 
	 * @param communicator
	 * @param jobTask
	 * @param request The complete request as a JSONObject
	 */
	public Job(ClientCommunicatorInterface communicator, String token, String jobTask, JSONObject request){
		this.jobId = getNextId();
		this.init(communicator, jobTask, token, request, "json");
	}
	
	private void init(ClientCommunicatorInterface communicator, String jobTask, String token, JSONObject details, String resultFormat){
		this.time_created = System.currentTimeMillis();
		this.communicator = communicator;
		this.jobTask = jobTask.toLowerCase();
		this.token = token;
		this.state = Job.STATE_NEW;
		this.request = details;
		this.resultFormat = resultFormat;
		this.statusMessage = null;
		this.statusCode = -1;
		this.error = null;
	}
	
	/**
	 * Returns the name of the task this job is intended to execute.
	 * 
	 * @return The task this job should execute.
	 */
	public String getJobTask(){
		return this.jobTask;
	}
	
	/**
	 * Returns the user token for this job.
	 * @return
	 */
	public String getToken(){
		return this.token;
	}
	
	public String toString(){
		return this.jobId+": "+this.jobTask+" - "+this.state+" - "+this.statusCode+" - "+this.statusMessage;
	}
	
	public String getTimes(){
		return "Total execution time: "+(this.time_completed-this.time_created)+"ms\n";
	}
}
