package core;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;

import org.json.JSONObject;

import core.errorHelper.ErrorObject;
import core.jobManagerExceptions.InternalErrorException;
import core.jobManagerExceptions.JobManagerNotRunningException;
import core.jobManagerExceptions.NoJobExecutorException;
import core.jobManagerExceptions.QueueIsFullException;
import core.userData.UserData;
import core.userData.exceptions.DatabaseConnectionException;
import core.userData.exceptions.TokenNotFoundException;

import jobExecutors.JobExecutorInterface;

/**
 * Handles management of job schedules
 * 
 * @author Thomas Rudfoss
 *
 */
public final class JobManager {

	protected boolean active; //True while the job manager can run, false when terminating. If false on check after initialization then call terminate
	
	private final int threadCount;
	private final int dbConnectionCount;
	private final int maxQueueSize;
	private final long maxJobLiveTime;
	
	private int finishedJobs;
	private int failedJobs;
	private int activeJobs;
	private int timedOutJobs;
	
	protected final int maxDBConnectionAttempts;
	
	private final LinkedList<Job> waitingWorkQueue;
	private final LinkedList<WorkerThread> waitingForDBConnections;
	private final LinkedList<DBConnection> freeDBConnections;
	
	private final WorkerThread[] workerThreads;
	private final DBConnection[] dbConnections;
	
	protected DBConnection coreDBCon;
	
	protected JobManager(){
		this.active = true;
		Core.logger.info("JobManager starting...");
		
		this.threadCount = Core.config.get("core.workerThreadCount").asInt();
		this.maxQueueSize = Core.config.get("core.maxJobQueueSize").asInt();
		this.dbConnectionCount = Core.config.get("core.db.workerDBConnectionCount").asInt();
		this.maxJobLiveTime = Core.config.get("core.jobTimeout").asLong()*1000;
		
		this.finishedJobs = 0;
		this.failedJobs = 0;
		this.activeJobs = 0;
		this.timedOutJobs = 0;
		this.maxDBConnectionAttempts = Core.config.get("core.db.connectionRetries").asInt();
		
		this.waitingWorkQueue = new LinkedList<Job>();
		this.waitingForDBConnections = new LinkedList<WorkerThread>();
		this.freeDBConnections = new LinkedList<DBConnection>();
		
		Core.logger.fine("JobManager initialized ("+this.threadCount+" threads, "+this.maxQueueSize+" max queue size, "+this.dbConnectionCount+" dbConnections)");
		
		this.workerThreads = new WorkerThread[this.threadCount];
		for ( int i=0; i<this.threadCount; i++ ){
			this.workerThreads[i] = new WorkerThread();
			this.workerThreads[i].start();
		}
		
		this.dbConnections = new DBConnection[this.dbConnectionCount];
		for ( int i=0; i<=this.dbConnectionCount; i++ ){
			Connection c = null;
			int attempts = 0;
			while ( attempts < this.maxDBConnectionAttempts ){				
				try {
					c = DriverManager.getConnection(Core.config.get("core.db.url").value(),
							Core.config.get("core.db.user").value(),
							Core.config.get("core.db.password").value());
					if ( !c.isClosed() ){
						break;
					}
				} catch ( SQLException e ){
					Core.logger.warning("JobManager: Failed to connect to database! Please check your configuration settings.");
					Core.logger.warning(Core.stackTraceToString(e));
					this.active = false;
					break;
				}
				
				attempts++;
			}
			
			try {
				if ( c != null && !c.isValid(0) ){
					Core.logger.warning("JobManager: Connection to database successful, but still failed somewhere!");
					this.active = false;
					break;
				}
			} catch (SQLException e) {
				Core.logger.warning("JobManager: Failed to check for active connection!");
				Core.logger.warning(Core.stackTraceToString(e));
				this.active = false;
				break;
			}
			
			if ( !this.active ) break;

			
			if ( i == this.dbConnectionCount ){
				//Create db connection for the core and associated classes
				Core.logger.info("JobManager: Core database connection created");
				coreDBCon = new DBConnection(c);
				break;
			}
			
			this.dbConnections[i] = new DBConnection(c);
			this.freeDBConnections.add(this.dbConnections[i]);
		}
	}

	/**
	 * Safely terminate the JobManager and all jobs currently running. Jobs that are queued, but have not begun executing will also be stopped.
	 */
	protected void terminate(){
		this.active = false;
		
		for ( int i=0; i<this.workerThreads.length; i++ ){
			this.workerThreads[i].safeStop();
		}
		
		for ( int i=0; i<this.dbConnections.length; i++ ){
			try {
				this.dbConnections[i].close();
			} catch (SQLException e) {
				Core.logger.warning("JobManager: Could not terminate database connection.");
				Core.logger.warning(Core.stackTraceToString(e));
			}
		}
		
		try {
			coreDBCon.close();
		} catch (SQLException e) {
			Core.logger.warning("JobManager: Could not terminate core database connection.");
			Core.logger.warning(Core.stackTraceToString(e));
		}
	}
	
	protected void printState(){
		System.out.println("System state per "+(new Date()).toString());
		System.out.println("Waiting jobs:                "+waitingWorkQueue.size()+"/"+maxQueueSize);
		System.out.println("Active jobs:                 "+activeJobs);
		System.out.println("Finished jobs:               "+finishedJobs);
		System.out.println("Failed jobs:                 "+failedJobs);
		System.out.println("Worker threads:              "+threadCount);
		System.out.println("Database connections:        "+dbConnectionCount);
	}
	
	protected synchronized void clearQueuedJobs(){
		Iterator<Job> it = this.waitingWorkQueue.iterator();
		while ( it.hasNext() ){
			Job job = it.next();
			job.state = Job.STATE_CANCELED;
			job.communicator.notifyMe();
		}
		
		this.waitingWorkQueue.clear();
	}
	
	protected boolean jobHasTimedOut(Job job){
		if ( (System.currentTimeMillis() - job.time_created) > this.maxJobLiveTime ){
			return true;
		}
		
		return false;
	}
	
	protected void translateResult(Job theJob){
		Core.logger.finer("JobExecutor: Translating job "+theJob+" to "+theJob.resultFormat);
		//Translate result
		if ( !(theJob.internalResult instanceof JSONObject) ){
			if ( theJob.internalResult == null ) Core.logger.warning("JobManager: "+theJob+" returned result of type: NULL");
			else Core.logger.warning("JobManager: "+theJob+" returned result of type: "+theJob.internalResult.getClass().getName());
			theJob.result = "JobExecutor returned invalid result format";
		} else {
			if ( theJob.resultFormat.equalsIgnoreCase("json") ){
				theJob.result = theJob.internalResult.toString();
			} else {
				//At this point we are sure the jobTranslator exists and can proceed accordingly. 
				ResultTranslatorWrapper.getTranslatorTo(theJob.resultFormat).translateJob(theJob, theJob.resultFormat);
			}
			
			theJob.state = Job.STATE_COMPLETE;
		}
	}
	
	
	
	
	
	
	
	/**
	 * A thread that performs the work in a job
	 * 
	 * @author Thomas Rudfoss
	 *
	 */
	private class WorkerThread extends Thread{
		
		private boolean run;
		
		public WorkerThread(){
			super();
			this.run = true;
			Core.logger.finer("JobManager: New WorkerThread created.");
		}
		
		public void run(){
			Job theJob = null;
			Core.logger.finer("JobManager: WorkerThread running.");
			
			while ( this.run ){
				boolean internalShutdown = false;
				
				synchronized(waitingWorkQueue){
					//Wait for a job that it can execute
					while ( waitingWorkQueue.isEmpty() ){
						Core.logger.finer("JobManager: Waiting for new job");
						try { waitingWorkQueue.wait(); }
						catch (InterruptedException e){
							internalShutdown = true;
							break;
						}
					}
					
					if ( internalShutdown ) break;
					theJob = waitingWorkQueue.removeFirst();
					theJob.time_dequeued = System.currentTimeMillis();
				}
				
				Core.logger.finer("JobManager: Job in queue retrieved.");
				activeJobs++;
				
				//Check that token exists
				try {
					UserData ud = UserData.constructFromToken(theJob.getToken());
					theJob.userData = ud;
				} catch ( DatabaseConnectionException e ){
					ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob, "Internal server error");
					Core.logger.warning("JobManager: Database error. Cannot get user from token.");
					translateResult(theJob);
					theJob.state = Job.STATE_FAILED;
					theJob.communicator.notifyMe();
					failedJobs++;
					
					activeJobs--;
					Core.logger.finer("JobManager: Worker thread finished. Returning to pool.");
					continue;
				} catch ( TokenNotFoundException e ){
					ErrorObject.getErrorForCode(ErrorObject.CODE_NO_SUCH_TOKEN, theJob, "No such token found");
					Core.logger.warning("JobManager: No user matching token found. Returning job");
					translateResult(theJob);
					theJob.state = Job.STATE_FAILED;
					theJob.communicator.notifyMe();
					failedJobs++;
					
					activeJobs--;
					Core.logger.finer("JobManager: Worker thread finished. Returning to pool.");
					continue;
				}
				
				while ( !requestDBConForJob(theJob) ){
					queueWorkerWaitingForDB(this);
					try {
						this.wait();
					} catch (InterruptedException e) {
						internalShutdown = true;
						break;
					}
				}
				
				if ( internalShutdown ) break;
				
				Core.logger.finer("JobManager: Job received database connection");
				
				try {
					if ( jobHasTimedOut(theJob) ){
						Core.logger.warning("JobManager: Job timed out. ID: "+theJob.jobId);
						theJob.state = Job.STATE_TIMEOUT;
						timedOutJobs++;
					} else {
						Core.logger.finer("JobManager: Job executing. ID: "+theJob.jobId);
						theJob.time_jobStart = System.currentTimeMillis();
						theJob.executor.doJob(theJob); //Perform the actual job
						theJob.time_jobFinished = System.currentTimeMillis();
						theJob.time_completed = System.currentTimeMillis();
						finishedJobs++;
						Core.logger.finer("JobManager: Job completed. ID: "+theJob.jobId);
					}
					
					if ( jobHasTimedOut(theJob) ){
						Core.logger.warning("JobManager: Job timed out. ID: "+theJob.jobId);
						theJob.state = Job.STATE_TIMEOUT;
						timedOutJobs++;
					}
					
					//Translate result
					translateResult(theJob);
					
					//Fix status for the job
					if ( theJob.state != Job.STATE_COMPLETE ){
						if ( theJob.state == Job.STATE_CANCELED ){
							theJob.statusCode = 60;
							theJob.statusMessage = "Cancelled";
							theJob.result = ResponseTemplateTypes.JOB_CANCELED.json;
						} else if ( theJob.state == Job.STATE_TIMEOUT ){
							theJob.statusCode = 50;
							theJob.statusMessage = "Tiemout";
							theJob.result = ResponseTemplateTypes.JOB_TIMEOUT.json;
						} else if ( theJob.state == Job.STATE_FAILED ){
							//The job executor should in these cases itself have set the result (statusCode, statusMessage and result) to one of the ResponseTemplateTypes
						}
					}
					
					workerThreadFinishedWithDB(theJob);
					long runtime = theJob.time_completed - theJob.time_created;
					long execTime = theJob.time_jobFinished - theJob.time_jobStart;
					long overheadTime = (theJob.time_completed-theJob.time_created)-execTime;
					String timeStr = "JobManager: Job finished (Runtime: "+runtime+"ms / Overhead: "+overheadTime+"ms / ExecutionTime: "+execTime+"ms)";
					if ( Core.config.get("logging.logTimers").asBoolean() ) Core.logger.finest(timeStr);
					System.out.println(timeStr);
					Core.logger.finer("JobManager: Notifying communicator "+theJob.communicator);
					theJob.communicator.notifyMe();
				} catch (RuntimeException e){
					failedJobs++;
					theJob.state = Job.STATE_FAILED;
					Core.logger.warning("JobManager: RuntimeException");
					Core.logger.warning(Core.stackTraceToString(e));
					Core.logger.finer("JobManager: Notifying communicator "+theJob.communicator);
					theJob.communicator.notifyMe();
				} catch (Exception e){
					failedJobs++;
					theJob.state = Job.STATE_FAILED;
					Core.logger.warning("JobManager: Exception");
					Core.logger.warning(Core.stackTraceToString(e));
					Core.logger.finer("JobManager: Notifying communicator "+theJob.communicator);
					theJob.communicator.notifyMe();
				}
				
				activeJobs--;
				Core.logger.finer("JobManager: Worker thread finished successfully. Returning to pool.");
			}
			
			Core.logger.finer("JobManager: WorkerThread terminated.");
		}
		
		/**
		 * Safely terminate the thread by stopping it and then calling interrupt.
		 * WARNING: JobExecutors should periodically check if the thread has been interrupted if they will support this.
		 * If not the thread will continue executing as if nothing has happened.
		 */
		public void safeStop(){
			this.run = false;
			this.interrupt();
		}
	}
	
	
	
	
	
	

	/**
	 * Add a new job to the end of the job queue.
	 * 
	 * @param job
	 * @throws JobManagerNotRunningException 
	 * @throws QueueIsFullException 
	 * @throws NoJobExecutorException 
	 * @throws ClassNotFoundException 
	 * @throws InstantiationException 
	 * @throws IllegalAccessException 
	 */
	protected void queueJob(Job job) throws JobManagerNotRunningException, QueueIsFullException, NoJobExecutorException, InternalErrorException{
		if ( !this.active ) throw new JobManagerNotRunningException();
		
		if ( this.waitingWorkQueue.size() >= this.maxQueueSize ) throw new QueueIsFullException();
		
		Class<?> jobExecutorClass = Core.jobExecutorList.get(job.getJobTask());
		if ( jobExecutorClass == null ) throw new NoJobExecutorException(job.getJobTask());//No JobExecutor found that can do this job
		
		try {
			job.executor = (JobExecutorInterface) jobExecutorClass.newInstance();
		} catch (InstantiationException e) {
			Core.logger.warning("JobManager: Could not create new instance of "+jobExecutorClass+" for a job. Instansiation exception!");
			Core.logger.warning(Core.stackTraceToString(e));
			throw new InternalErrorException();
		} catch (IllegalAccessException e) {
			Core.logger.warning("JobManager: Could not create new instance of "+jobExecutorClass+" for a job. Access denied when instanciating");
			throw new InternalErrorException();
		}
		
		//Check if a translator for the format exists
		if ( (!job.resultFormat.equalsIgnoreCase("json")) && (!ResultTranslatorWrapper.translatorExistsFor(job.resultFormat)) ){
			job.state = Job.STATE_FAILED;
			job.statusCode = 44;
			job.statusMessage = "No translator for that format";
			Core.logger.warning("JobManager: No translator found for the format: "+job.resultFormat);
			throw new InternalErrorException();
		}
		
		synchronized(this.waitingWorkQueue){
			Core.logger.finer("JobManager: Notifying of new job");
			job.time_queued = System.currentTimeMillis();
			this.waitingWorkQueue.add(job);
			this.waitingWorkQueue.notify();
		}
	}
	
	private boolean requestDBConForJob(Job job){
		if ( this.freeDBConnections.size() > 0 ){
			job.dbConnection = this.freeDBConnections.removeFirst();
			return true;
		}
		
		return false;
	}
	
	private void queueWorkerWaitingForDB(WorkerThread wt){
		this.waitingForDBConnections.add(wt);
	}
	
	private void workerThreadFinishedWithDB(Job job) throws SQLException{
		job.dbConnection.clean();
		this.freeDBConnections.add(job.dbConnection);
		job.dbConnection = null;
		job.state = Job.STATE_COMPLETE;
		
		if ( this.waitingForDBConnections.size() > 0 ){
			WorkerThread wt = this.waitingForDBConnections.removeFirst();
			synchronized (wt){
				wt.notify();
			}
		}
	}
}
