package core;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import resultTranslators.ResultTranslatorInterface;


/**
 * This class holds the necessary information for the core to use resultTranslator.
 * This class is used by the core and should not be used in any plugin.
 * 
 * A resultTranslater can translate from JSON (the internal format) to another format.
 * 
 * Any ResultTranslator
 * 
 * @author Thomas Rudfoss
 *
 */
public class ResultTranslatorWrapper {
	
	/**
	 * Holds a reference to the resultTranslator class that can translate to a given format.
	 * 
	 * xml-IMS -> toXMLIMSformat (reference)
	 */
	private static HashMap<String, ResultTranslatorWrapper> translators  = new HashMap<String, ResultTranslatorWrapper>();
	
	/**
	 * Add a new resultTranslator plugin to the translators list.
	 * 
	 * @param resultTranslatorPlugin The class that implements a given translator.
	 * @return True if successfully added. False if not.
	 */
	protected static boolean addResultTranslator(Class<?> resultTranslatorPlugin){
		String[] toFormats = new String[0];
		ResultTranslatorWrapper tranWrap = null;
		
		try {
			toFormats = (String[]) resultTranslatorPlugin.getMethod("iCanTranslateTo").invoke((Object)null);
			
			tranWrap = new ResultTranslatorWrapper(resultTranslatorPlugin);
			for ( int i=0; i<toFormats.length; i++ ){
				String toFormat = toFormats[i].toLowerCase();
				translators.put(toFormat, tranWrap);
				Core.logger.fine("ResultTranslatorWrapper: Added translator for "+toFormat);
			}
		} catch (IllegalArgumentException e) {
			Core.logger.warning("ResultTranslatorWrapper: Could not wrap "+resultTranslatorPlugin.getName()+"");
			Core.logger.warning(Core.stackTraceToString(e));
			return false;
		} catch (SecurityException e) {
			Core.logger.warning("ResultTranslatorWrapper: Could not wrap "+resultTranslatorPlugin.getName()+"");
			Core.logger.warning(Core.stackTraceToString(e));
			e.printStackTrace();
			return false;
		} catch (IllegalAccessException e) {
			Core.logger.warning("ResultTranslatorWrapper: Could not wrap "+resultTranslatorPlugin.getName()+"");
			Core.logger.warning(Core.stackTraceToString(e));
			e.printStackTrace();
			return false;
		} catch (InvocationTargetException e) {
			Core.logger.warning("ResultTranslatorWrapper: Could not wrap "+resultTranslatorPlugin.getName()+"");
			Core.logger.warning(Core.stackTraceToString(e));
			e.printStackTrace();
			return false;
		} catch (NoSuchMethodException e) {
			Core.logger.warning("ResultTranslatorWrapper: Could not wrap "+resultTranslatorPlugin.getName()+"");
			Core.logger.warning(Core.stackTraceToString(e));
			e.printStackTrace();
			return false;
		}
		
		return true;
	}
	
	/**
	 * Returns a translator instance that will translate from the fromFormat to toFormat if such a translator exists.
	 * If not then null is returned.
	 * 
	 * @param fromFormat
	 * @param toFormat
	 * @return
	 */
	protected static ResultTranslatorInterface getTranslatorTo(String toFormat){
		if ( translators.containsKey(toFormat) ) return translators.get(toFormat).getTranslator();
		
		return null;
	}
	
	protected static boolean translatorExistsFor(String format){
		if ( ResultTranslatorWrapper.getTranslatorTo(format) == null ) return false;
		
		return true;
	}
	
	
	private ArrayList<String> toFormats;
	private Class<?> resultTranslatorPlugin;
	
	/**
	 * Create a new wrapper for a resultTranslator plugin. Once complete this wrapper can be used for easy access to the translator.
	 * 
	 * @param resultTranslatorPlugin
	 */
	protected ResultTranslatorWrapper(Class<?> resultTranslatorPlugin){
		this.resultTranslatorPlugin = resultTranslatorPlugin;
	}
	
	/**
	 * Returns a valid instance of the resultTranslator it contains or null if an instance could not be created.
	 * 
	 * @return
	 */
	protected ResultTranslatorInterface getTranslator(){
		try {
			return (ResultTranslatorInterface) resultTranslatorPlugin.newInstance();
		} catch (InstantiationException e) {
			Core.getCore().getPluginLogger().warning("ResultTranslatorWrapper: InstantiationException occured on "+this.resultTranslatorPlugin.getClass().getName());
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
			return null;
		} catch (IllegalAccessException e) {
			Core.getCore().getPluginLogger().warning("ResultTranslatorWrapper: IllegalAccessException occured on "+this.resultTranslatorPlugin.getClass().getName());
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
			return null;
		}
	}
	
	/**
	 * Checks if the resultTranslator can translate to the requested format.
	 * This method should be case-insensitive.
	 * 
	 * @param toFormat The format to check (xml, json, resultset)
	 * @return
	 */
	protected boolean canTranslateTo(String toFormat){
		Iterator<String> it = this.toFormats.iterator();
		while (it.hasNext()){
			if ( it.next().equalsIgnoreCase(toFormat) ) return true;
		}
		
		return false;
	}

}
