package core.errorHelper;

import org.json.JSONException;
import org.json.JSONObject;

import core.Core;
import core.Job;

public class ErrorObject {
	
	/**
	 * No error. Everything is fine.
	 */
	public static final int CODE_SUCCESSFULL = 0;
	/**
	 * The token identifies a user that does not have access to the requested operation
	 */
	public static final int CODE_ACCESS_DENIED = 10;
	/**
	 * There is no user with this token
	 */
	public static final int CODE_NO_SUCH_TOKEN = 20;
	/**
	 * The format of the request is invalid (i.e. got JSON when expecting XML)
	 */
	public static final int CODE_INVALID_REQUEST_FORMAT = 30;
	/**
	 * Some data is missing from the request
	 */
	public static final int CODE_MISSING_DATA = 35;
	/**
	 * Server is too busy to process the request
	 */
	public static final int CODE_SERVER_BUSY = 40;
	/**
	 * There is no job executor that can to this job
	 */
	public static final int CODE_NO_JOBEXECUTOR = 42;
	/**
	 * There is no translator that can translate to this format
	 */
	public static final int CODE_NO_TRANSLATOR = 44;
	/**
	 * The job timed out before it could complete
	 */
	public static final int CODE_TIMEOUT = 50;
	/**
	 * The job was canceled before it began executing
	 */
	public static final int CODE_CANCELED = 60;
	/**
	 * An error occurred on the server that the user may not need to know anything about 
	 */
	public static final int CODE_SERVER_ERROR = 100;
	/**
	 * An error that is unknown to the core, but probably known to the module that caused it
	 */
	public static final int CODE_UNKNOWN_ERROR = 200;

	/**
	 * Populates the given job with the proper values and status to describe the error code given.
	 * 
	 * @param code One of the constants defined in this class (CODE).
	 * @param job
	 * @return
	 */
	public static ErrorObject getErrorForCode(int code, Job job){
		String message = null;
		switch ( code ){
		case 10:
			message = "Access denied";
			break;
		case 20:
			message = "No such token exists";
			break;
		case 30:
			message = "The request format is not valid";
			break;
		case 35:
			message = "The request is missing some data required to perform the job";
			break;
		case 40:
			message = "The server is too busy to process your request";
			break;
		case 42:
			message = "There is no job executor for that job";
			break;
		case 44:
			message = "There is no translator for that result format";
			break;
		case 50:
			message = "The job has timed out before completing";
			break;
		case 60:
			message = "The job was canceled before it could begin executing";
			break;
		case 100:
			message = "Internal server error";
			break;
		case 200:
			message = "Unknown error";
			break;
		default:
			message = "Unknown server error";
			break;
		}
		return getErrorForCode(code, job, message);
	}
	
	public static ErrorObject getErrorForCode(int code, Job job, String message){
		switch ( code ){
		case 50:
			job.state = Job.STATE_TIMEOUT;
			break;
		case 60:
			job.state = Job.STATE_CANCELED;
			break;
		default:
			job.state = Job.STATE_FAILED;
			break;
		}
		
		return getErrorForCode(code, job, message, 0);
	}
	
	public static ErrorObject getErrorForCode(int code, Job job, String message, int internalErrorCode){
		switch ( code ){
		case 50:
			job.state = Job.STATE_TIMEOUT;
			break;
		case 60:
			job.state = Job.STATE_CANCELED;
			break;
		default:
			job.state = Job.STATE_FAILED;
			break;
		}
		
		return new ErrorObject(code, job, message, internalErrorCode);
	}
	
	
	
	
	

	public final int jobStatusCode;
	public final String jobStatusMessage;
	public final String jobResult;
	public final int internalErrorCode;
	public final Job theJob;
	
	public ErrorObject(int jobStatusCode, Job theJob, String jobStatusMessage, int internalErrorCode){
		this.jobStatusCode = jobStatusCode;
		this.jobStatusMessage = jobStatusMessage;
		this.internalErrorCode = internalErrorCode;
		this.theJob = theJob;
		this.jobResult = this.getJSONResult();
		Core.getCore().getPluginLogger().finer("ErrorObject: ResultJSON: "+this.jobResult);
		
		this.theJob.statusCode = this.jobStatusCode;
		this.theJob.statusMessage = this.jobStatusMessage;
		
		try {
			this.theJob.internalResult = new JSONObject(this.jobResult);
		} catch (JSONException e) {
			Core.getCore().getPluginLogger().severe("ErrorObject: Could not generate JSONObject for internal result.");
			Core.getCore().getPluginLogger().severe(Core.stackTraceToString(e));
		}
		
		this.theJob.error = this;
	}
	
	private String getJSONResult(){
		return "{\"response\":{\"responsespec\":{\"status\":"+this.jobStatusCode+",\"errcode\":"+this.internalErrorCode+",\"statusmessage\":\""+jobStatusMessage+"\"}}}";
	}
}
