package core.userData;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * Contains a list of values for a given option.
 * A role can have multiple values in several cases for instance when defining course participation.
 * 
 * @author Thomas Rudfoss
 *
 */
public class RoleOptionList {
		
	private String optionName;
	private ArrayList<String> optionValues;
	
	public RoleOptionList(String optionName){
		this.optionName = optionName;
		this.optionValues = new ArrayList<String>();
	}
	
	/**
	 * Return the name of the option this list contains values for.
	 * 
	 * @return
	 */
	public String getName(){
		return this.optionName;
	}
	
	/**
	 * Add a value to this list (used during construction and should not be used afterwards)
	 * 
	 * @param value
	 */
	protected void addValue(String value){
		optionValues.add(value);
	}
	
	/**
	 * Checks if the requested value is in the list of values (case-insensitive).
	 * 
	 * @param value
	 * @return
	 */
	public boolean hasValue(String value){
		Iterator<String> it = optionValues.iterator();
		while ( it.hasNext() ){
			if ( it.next().equalsIgnoreCase(value) ) return true;
		}
		
		return false;
	}
	
	/**
	 * Returns all values in the list.
	 * 
	 * @return
	 */
	public ArrayList<String> getValues(){
		return this.optionValues;
	}
	
	/**
	 * Returns the first value in the list as a string.
	 * 
	 * @return
	 */
	public String getString(){
		return getString(0);
	}
	
	/**
	 * Returns a specific value in the list based on the index as a string.
	 * 
	 * @param index
	 * @return
	 */
	public String getString(int index){
		if ( index >= optionValues.size() ){
			return null;
		}
		
		return optionValues.get(index);
	}
	
	/**
	 * Returns the first value in the list as an integer if possible.
	 * 
	 * @return
	 */
	public Integer getInt(){
		return getInt(0);
	}
	
	/**
	 * Returns a specific value in the list based on the index as an integer or null if out of bounds.
	 * 
	 * @param index
	 * @return
	 */
	public Integer getInt(int index){
		if ( index >= optionValues.size() ){
			return null;
		}
		
		return Integer.parseInt(optionValues.get(index));
	}
	
	/**
	 * Returns the first value in the list as a double precision number (double) if possible.
	 * 
	 * @return
	 */
	public Double getDouble(){
		return getDouble(0);
	}
	
	/**
	 * Returns a specific value in the list based on the index as a double or null if out of bounds.
	 * 
	 * @param index
	 * @return
	 */
	public Double getDouble(int index){
		if ( index >= optionValues.size() ){
			return null;
		}
		
		return Double.parseDouble(optionValues.get(index));
	}
	
	/**
	 * Returns the first value in the list as a boolean value.
	 * The following values are interpreted as true (case-insensitive):
	 * yes, true, 1
	 * 
	 * @return
	 */
	public Boolean getBool(){
		return getBool(0);
	}
	
	/**
	 * Returns a specific value in the list based on the index as a boolean value or null if out of bounds.
	 * The following values are interpreted as true (case-insensitive):
	 * yes, true, 1
	 * 
	 * @param index
	 * @return
	 */
	public Boolean getBool(int index){
		if ( index >= optionValues.size() ){
			return null;
		}

		String value = optionValues.get(index);
		if ( value.equalsIgnoreCase("yes") ) return true;
		if ( value.equalsIgnoreCase("true") ) return true;
		if ( value.equalsIgnoreCase("1") ) return true;
		
		return false;
	}
	
	/**
	 * Returns a string representation of this RoleOptionList with the name and all values. 
	 */
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append(optionName).append("\n");
		Iterator<String> it = optionValues.iterator();
		while ( it.hasNext() ){
			sb.append("->"+it.next());
		}
		
		return sb.toString();
	}
}
