package core.userData;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;

import core.Core;
import core.DBConnection;
import core.userData.exceptions.DatabaseConnectionException;
import core.userData.exceptions.NoSuchRoleException;
import core.userData.exceptions.TokenNotFoundException;
import core.userData.exceptions.UserIdNotFoundException;

/**
 * Class containing user specific functionality for identification and authentication of users based on tokens.
 * 
 * Once created a UserData object cannot be modified.
 * 
 * 
 * @author Thomas Rudfoss
 *
 */
public class UserData {
		
	private static HashMap<String, PreparedStatement> queries;
	
	private static boolean ready = false;
	
	/**
	 * Sets the core database connection if it is not already set.
	 * 
	 * @param db
	 */
	public static void initUserData(DBConnection db){
		if ( db == null ) return;
		
		queries = new HashMap<String, PreparedStatement>();
		String[] rawQueries = new String[]{
				"SELECT identifier FROM User WHERE token=? LIMIT 1",
				"SELECT ar.name, ri.identifier, uri.assignedDateTime, uri.expireDateTime FROM AccessRole as ar, RoleInstance as ri, UserRoleInstance as uri WHERE uri.user_identifier=? AND ri.identifier=uri.roleInstance_identifier AND ar.identifier=ri.accessRole_identifier",
				"SELECT ro.optionName, rov.optionValue FROM RoleOption as ro, RoleOptionValue as rov WHERE rov.roleInstance_identifier=? AND ro.identifier=rov.roleOption_identifier",
				"SELECT token FROM User WHERE identifier=?"
			};
				
		//Initialize queries:
		try {
			queries.put("getUserIdentifierForToken", db.newPreparedStatement(rawQueries[0]));
			queries.put("getUserRoleInstanceForUser", db.newPreparedStatement(rawQueries[1]));
			queries.put("getRoleOptionsForRoleInstance", db.newPreparedStatement(rawQueries[2]));
			queries.put("getUserTokenForId", db.newPreparedStatement(rawQueries[3]));
			
		} catch (SQLException e) {
			Core.getCore().getPluginLogger().severe("UserData: Could not prepare query. UserData is not accessible");
			Core.getCore().getPluginLogger().severe(Core.stackTraceToString(e));
			return;
		}
		
		ready = true;
	}
	
	/**
	 * Constructs a new UserData instance based on a users id.
	 * 
	 * @param id The id of the user element in the table.
	 * @return
	 * @throws DatabaseConnectionException
	 * @throws UserIdNotFoundException
	 */
	public static UserData constructFromId(int id) throws DatabaseConnectionException, UserIdNotFoundException{
		String token = "";
		
		try {
			PreparedStatement stmt = queries.get("getUserTokenForId");
			stmt.setInt(1, id);
			stmt.execute();
			if ( !stmt.getResultSet().next() ){ throw new UserIdNotFoundException(Integer.toString(id));}
			token = stmt.getResultSet().getString(1);
		} catch ( SQLException e ){
			Core.getCore().getPluginLogger().severe("UserData: Could not execute query on server. Cannot create UserData object");
			throw new DatabaseConnectionException(Core.stackTraceToString(e));
		}
		
		try {
			return constructFromToken(token);
		} catch ( TokenNotFoundException e ){
			Core.getCore().getPluginLogger().severe("UserData: Attempted to construct from id that is known to exists but got TokenNotFoundException!");
			Core.getCore().getPluginLogger().severe(Core.stackTraceToString(e));
		}
		
		return null;
	}
	
	/**
	 * Constructs a new UserData instance based on a users token.
	 * 
	 * @param token
	 * @return
	 * @throws DatabaseConnectionException
	 * @throws TokenNotFoundException
	 */
	public static UserData constructFromToken(String token) throws DatabaseConnectionException, TokenNotFoundException{
		if ( !ready ) return null;
		
		UserData user = null;
		
		try {
			//Get user ID in the system
			PreparedStatement uId = queries.get("getUserIdentifierForToken");
			uId.setString(1, token);
			ResultSet rs = uId.executeQuery();
			if ( !rs.next() ) throw new TokenNotFoundException(token);
			int userId = rs.getInt(1);
			
			user = new UserData(token, userId);
			
			//Get all roles for that user
			PreparedStatement roleQuery = queries.get("getUserRoleInstanceForUser");
			roleQuery.setInt(1, userId);
			rs = roleQuery.executeQuery();
			while ( rs.next() ){
				//One or more roles found
				String roleName = rs.getString(1);
				int roleId = rs.getInt(2);
				Date assigned = UserData.makeDateFromSQL(rs.getString(3));
				Date expires = null;
				String expireTime = rs.getString(4);
				if ( !rs.wasNull() ) expires = UserData.makeDateFromSQL(expireTime);
								
				UserRole aRole = new UserRole(roleName, assigned, expires);
				
				//Grab all role options if there are any
				PreparedStatement roleOptions = queries.get("getRoleOptionsForRoleInstance");
				roleOptions.setInt(1, roleId);
				ResultSet roleOpRs = roleOptions.executeQuery();
				while ( roleOpRs.next() ){
					String optionName = roleOpRs.getString(1);
					String optionValue = roleOpRs.getString(2);
					
					if ( aRole.hasOption(optionName) ){
						aRole.getOption(optionName).addValue(optionValue);
					} else {
						RoleOptionList rol = new RoleOptionList(optionName);
						rol.addValue(optionValue);
						aRole.setOption(rol);
					}
				}
				
				user.setRole(roleName, aRole);
			}
			
			//UserData created
		} catch (SQLException e) {
			Core.getCore().getPluginLogger().severe("UserData: Could not execute query on server. Cannot create UserData object");
			throw new DatabaseConnectionException(Core.stackTraceToString(e));
		}
		
		return user;
	}
	
	/**
	 * Constructs a date object from an SQL date string (yyyy-mm-dd hh:mm:ss)
	 * 
	 * @param sqlDate
	 * @return
	 */
	public static Date makeDateFromSQL(String sqlDate){
		//Makes a valid date object from an sql string yyyy-mm-dd hh:mm:ss
		Calendar cal = new GregorianCalendar(Integer.parseInt(sqlDate.substring(0, 4), 10),
				Integer.parseInt(sqlDate.substring(5, 7), 10)-1, 
				Integer.parseInt(sqlDate.substring(8, 10), 10),
				Integer.parseInt(sqlDate.substring(11, 13), 10),
				Integer.parseInt(sqlDate.substring(14, 16), 10),
				Integer.parseInt(sqlDate.substring(17, 19), 10));
		cal.set(Calendar.MILLISECOND, 0);
		
		return cal.getTime();
	}
	
	
	
	
	
	private int id;
	private String token;
	private HashMap<String, UserRole> roles;
	
	
	
	
	
	private UserData(String token, int id){
		this.token = token;
		this.roles = new HashMap<String, UserRole>();
		this.id = id;
	}
	
	/**
	 * Give the user a specific role.
	 * 
	 * @param roleName The name of the role
	 * @param role The UserRole object containing details about this role.
	 */
	protected void setRole(String roleName, UserRole role){
		this.roles.put(roleName, role);
	}
	
	/**
	 * Returns the token associated with this UserData object.
	 * 
	 * @return
	 */
	public String getToken(){
		return this.token;
	}
	
	/**
	 * Returns the user id associated with this UserData object.
	 * 
	 * @return
	 */
	public int getId(){
		return this.id;
	}
	
	/**
	 * Checks if the user has a specific role (case-sensitive).
	 * 
	 * @param roleName The name of the role
	 * @return
	 */
	public boolean hasRole(String roleName){
		return roles.containsKey(roleName);
	}
	
	/**
	 * Returns a UserRole object describing a specific role if it exists.
	 * 
	 * @param roleName Name of the role (case-sensitive).
	 * @return
	 * @throws NoSuchRoleException
	 */
	public UserRole getRole(String roleName) throws NoSuchRoleException{
		if ( !this.hasRole(roleName) ) throw new NoSuchRoleException(roleName);
		return roles.get(roleName);
	}
	
	/**
	 * Checks if the user has administration rights to the system.
	 * 
	 * @return
	 */
	public boolean isAdmin(){
		if ( !this.hasRole(Core.getCore().getConfig().get("core.adminRoleName").val()) ) return false;
		UserRole admRole = this.getRole(Core.getCore().getConfig().get("core.adminRoleName").val());
		
		return admRole.isActive();
	}

	/**
	 * Returns a string representation of this UserData object with all roles, their options and values.
	 */
	public String toString(){
		String res = new String();
		
		res+="UserData object for token: "+this.token+"\r\n";
		if ( this.isAdmin() ) res+="This user is an admin!\r\n";
		res+="RoleCount: "+this.roles.size()+"\r\n";
		if ( this.roles.size() > 0 ) {
			Iterator<String> it = this.roles.keySet().iterator();
			while (it.hasNext()){
				String nextRoleName = it.next();
				res+=this.roles.get(nextRoleName).toString()+"\r\n";
			}
		}
		
		return res;
	}
}
