package core.userData;

import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

import core.userData.exceptions.NoSuchRoleOptionException;

/**
 * Holds information about a specific instance of a role.
 * Once created a UserRole cannot be modified.
 * 
 * @author Thomas Rudfoss
 *
 */
public class UserRole {
	
	private String roleName;
	private HashMap<String, RoleOptionList> roleOptions;
	private Date assigned;
	private Date expires; //Can be null
	
	/**
	 * 
	 * @param roleName
	 * @param assigned A date object defining when
	 * @param expires A date when the role expires. Can be null for roles that don't expire.
	 */
	public UserRole(String roleName, Date assigned, Date expires){
		this.roleName = roleName;
		this.roleOptions = new HashMap<String, RoleOptionList>();
		
		this.assigned = assigned;
		this.expires = expires;
	}
	
	/**
	 * Returns the name of the role.
	 * 
	 * @return Name of role
	 */
	public String getName(){
		return this.roleName;
	}
	
	/**
	 * Set an option to a specific value. Overwrites existing option if it exists.
	 * 
	 * @param optionName
	 * @param optionValue
	 */
	protected void setOption(RoleOptionList list){
		roleOptions.put(list.getName(), list);
	}
	
	/**
	 * Checks if the role has an option with the given name.
	 * 
	 * @param optionName Name of the option.
	 * @return True/false
	 */
	public boolean hasOption(String optionName){
		return roleOptions.containsKey(optionName);
	}
	
	/**
	 * Checks if this role has an option with the given value.
	 * It uses the method hasValue in RoleOptionList and is therefore case-insensitive on the value, but not the name.
	 * 
	 * @param optionName
	 * @param value
	 * @return
	 */
	public boolean hasOptionWithValue(String optionName, String value){
		if ( !this.hasOption(optionName) ) return false;
		
		if ( !this.getOption(optionName).hasValue(value) ) return false;
		
		return true;
	}
	
	/**
	 * Checks only if the role is expired. Use isActive() to ensure the role is actually active.
	 * 
	 * @return
	 */
	public boolean isExpired(){
		if ( this.expires == null ) return false;
		
		Date now = new Date();
		if ( now.getTime() < expires.getTime() ) return false;
		
		return true;
	}
	
	/**
	 * Checks if the role is active by checking both assigned date and expired date.
	 * Criteria for an active role:
	 * - Assignment date in the past AND
	 * - Expires date in the future
	 * 
	 * @return
	 */
	public boolean isActive(){
		if ( this.isExpired() ) return false;
		
		Date now = new Date();
		if ( now.getTime() < this.assigned.getTime() ) return false;
		
		return true;
	}
	
	/**
	 * Returns the value of the option by name as a string.
	 * 
	 * @param optionName Name of the option.
	 * @return
	 */
	public RoleOptionList getOption(String optionName){
		if ( !this.hasOption(optionName) ) throw new NoSuchRoleOptionException(optionName);
		
		return roleOptions.get(optionName);
	}

	/**
	 * Returns a string representation of this UserRole with all options and values.
	 */
	public String toString(){
		String res = new String();
		
		res+="->RoleName: "+this.roleName+"\r\n";
		res+="  OptionCount: "+this.roleOptions.size()+"\r\n";
		if ( this.roleOptions.size() > 0 ){
			Iterator<String> it = this.roleOptions.keySet().iterator();
			while ( it.hasNext() ){
				String opName = it.next();
				res+="    "+opName+" = "+this.roleOptions.get(opName);
			}
		}
		
		return res;
	}

}
