package jobExecutors;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Level;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import core.Core;
import core.Job;
import core.errorHelper.ErrorObject;

/**
 * Handles some person related operations
 * 
 * Testing complete: OK
 * 
 * @author Thomas Rudfoss
 *
 */
public class PersonJobExecutor implements JobExecutorInterface {
	
	public static final int RETURNCODE_NO_RESULT = 10;
	public static final int RETURNCODE_ELEMENT_EXISTS = 20;
	
	private static String[] jobsICanDoList = new String[]{"createPerson", "createPersonRole", "listPersons", "listPersonRoles", "createPersonRoleInstance"}; 

	public static String[] jobsICanDo(){
		return jobsICanDoList.clone();
	}
	
	private Job theJob = null;
	private JSONObject data = null;
	private JSONObject details = null;
	
	@Override
	public void doJob(Job job) {
		log("Going to work!");
		
		theJob = job;
		try {
			theJob.request = theJob.request.getJSONObject("request");
		} catch (JSONException e1) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_INVALID_REQUEST_FORMAT, theJob);
			log(Level.FINE, "Could not grab request node.");
			logStackTrace(Level.FINE, e1);
			return;
		}
		
		if ( !hasProperRights() ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_ACCESS_DENIED, theJob);
			return;
		}
		
		try {
			theJob.internalResult = new JSONObject("{\"response\":{\"responsespec\":{\"status\":0,\"statusmessage\":\"OK\"}}}");
		} catch (JSONException e) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
			log(Level.SEVERE, "Could not initialize result object!");
			logStackTrace(Level.SEVERE, e);
			return;
		}
		
		try {
			if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[0]) ){
				data = theJob.request.getJSONObject("data");
				this.createPerson();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[1]) ){
				data = theJob.request.getJSONObject("data");
				this.createPersonRole();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[2]) ){
				details = theJob.request.getJSONObject("details");
				this.listPersons();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[3]) ){
				this.listPersonRoles();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[4]) ){
				details = theJob.request.getJSONObject("details");
				data = theJob.request.getJSONObject("data");
				this.createPersonRoleInstance();
			}
			
			theJob.state = Job.STATE_RESULT_SET;
		} catch ( JSONException e ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_MISSING_DATA, theJob);
			log("Missing data for the request.");
			logStackTrace(e);
		} catch (SQLException e) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
			log(Level.WARNING, "SQL Exception occured when inserting into person.");
			logStackTrace(Level.WARNING, e);
		}
		
		log("Finished work!");
	}
	
	private void createPerson() throws JSONException, SQLException{
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO Person (firstname, lastname) VALUES (?, ?)");
		stmt.setString(1, data.getString("firstname"));
		stmt.setString(2, data.getString("lastname"));
		stmt.execute();
	}
	
	private void createPersonRole() throws JSONException, SQLException{
		//First check that the role table exists if not return error
		try {
			PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT * FROM "+data.getString("roleTableName")+" LIMIT 1");
			stmt.execute();
		} catch ( SQLException e ){
			//This exception should occur only if the table does not exist
			ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No person-role table with the name "+data.getString("roleTableName")+" exists. This must be created first.");
			log("Tried to create person-role with table "+data.getString("roleTableName")+" which doesn't exist.");
			return;
		}
		
		//Then check that a role with that name does not already exist
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT name FROM Role WHERE name=? LIMIT 1");
		stmt.setString(1, data.getString("name"));
		stmt.execute();
		int count = 0;
		ResultSet rs = stmt.getResultSet();
		if ( rs != null ){
			while ( rs.next() ){
				count++;
			}
			
			if ( count > 0 ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "The role "+data.getString("name")+" already exists", PersonJobExecutor.RETURNCODE_ELEMENT_EXISTS);
				return;
			}
		}
	
		stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO Role (name, roleTableName) VALUES (?, ?)");
		stmt.setString(1, data.getString("name"));
		stmt.setString(2, data.getString("roleTableName"));
		stmt.execute();
	}
	
	private void listPersons() throws JSONException, SQLException{
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT * FROM Person");
		stmt.execute();
		
		ResultSet rs = stmt.getResultSet();
		JSONArray dataRows = new JSONArray();
		theJob.internalResult.getJSONObject("response").put("data", (new JSONObject()).put("row", dataRows));
		if ( rs == null ){ //No result
			return;
		}
		
		if ( details.getBoolean("includeRoles") ){
			PreparedStatement roleList = theJob.dbConnection.newPreparedStatement("SELECT roleTableName FROM Role");
			roleList.execute();
			
			ArrayList<String> roleListNames = new ArrayList<String>();
			ResultSet roleListResultSet = roleList.getResultSet();
			if ( roleListResultSet != null ){
				while ( roleListResultSet.next() ){
					roleListNames.add(roleListResultSet.getString(1));
				}
			}
			
			while ( rs.next() ){
				JSONObject thisObj = new JSONObject();
				int personIdentifier = rs.getInt("identifier");
				for ( int i=1; i<=rs.getMetaData().getColumnCount(); i++ ){
					String colName = rs.getMetaData().getColumnName(i);
					
					thisObj.put(colName, rs.getObject(i));
					
					//Lookup roles
					Iterator<String> it = roleListNames.iterator();
					JSONArray roles = new JSONArray();
					while ( it.hasNext() ){
						JSONObject roleContainer = new JSONObject();
						
						String tableName = it.next();
						PreparedStatement roleTableLookup = theJob.dbConnection.newPreparedStatement("SELECT * FROM "+tableName+" WHERE person_identifier=?");
						roleTableLookup.setInt(1, personIdentifier);
						roleTableLookup.execute();
						ResultSet roleResult = roleTableLookup.getResultSet();
						if ( roleResult == null ) continue;
						
						while ( roleResult.next() ){
							JSONObject roleObj = new JSONObject();
							for ( int u=1; u<=roleResult.getMetaData().getColumnCount(); u++ ){
								String roleColName = roleResult.getMetaData().getColumnName(u);
								
								roleObj.put(roleColName, roleResult.getObject(u));
							}
						
							roleContainer.put(tableName, roleObj);
						}
						
						if ( roleContainer.length() > 0 ) roles.put(roleContainer);
					}
					
					thisObj.put("roles", roles);
				}
				
				dataRows.put(thisObj);
			}
		} else {
			while ( rs.next() ){
				JSONObject thisObj = new JSONObject();
				for ( int i=1; i<=rs.getMetaData().getColumnCount(); i++ ){
					String colName = rs.getMetaData().getColumnName(i);
					
					thisObj.put(colName, rs.getObject(i));
				}
				
				dataRows.put(thisObj);
			}
		}
	}
	
	private void listPersonRoles() throws JSONException, SQLException{
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT name, roleTableName FROM Role");
		stmt.execute();
		
		ResultSet rs = stmt.getResultSet();
		JSONArray dataRows = new JSONArray();
		theJob.internalResult.getJSONObject("response").put("data", (new JSONObject()).put("row", dataRows));
		
		if ( rs == null ){
			return;
		}
		
		while ( rs.next() ){
			JSONObject thisObj = new JSONObject();
			
			thisObj.put("name", rs.getString(1));
			thisObj.put("roleTableName", rs.getString(2));
			
			dataRows.put(thisObj);
		}
	}
	
	private void createPersonRoleInstance() throws JSONException, SQLException{
		//First check that the role table exists if not return error
		try {
			PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT * FROM "+details.getString("roleTableName")+" LIMIT 1");
			stmt.execute();
		} catch ( SQLException e ){
			//This exception should occur only if the table does not exist
			ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No person-role table with the name "+data.getString("roleTableName")+" exists. This must be created first.");
			log("Tried to create person-role instance in table "+data.getString("roleTableName")+" which doesn't exist.");
			return;
		}
		
		//Create an entry in the specified person-role table
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO "+details.getString("roleTableName")+" (person_identifier) VALUES (?)");
		stmt.setInt(1, data.getInt("person_identifier"));
		stmt.execute();
	}
	
	private boolean hasProperRights(){
		if ( theJob.userData.isAdmin() ) return true; //Admins always have access
		
		if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[0]) ){
			if ( theJob.userData.hasRole("personAdmin") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[1]) ){
			if ( theJob.userData.hasRole("personAdmin") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[2]) ){
			if ( theJob.userData.hasRole("personAdmin") ) return true;
			if ( theJob.userData.hasRole("viewAll") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[3]) ){
			if ( theJob.userData.hasRole("personAdmin") ) return true;
			if ( theJob.userData.hasRole("viewAll") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[4]) ){
			if ( theJob.userData.hasRole("personAdmin") ) return true;
		}
		
		return false;
	}
	
	private void log(Level level, String msg){
		Core.getCore().getPluginLogger().log(level, "PersonJobExecutor: "+msg);
	}
	
	private void logStackTrace(Level level, Exception e){
		log(level, Core.stackTraceToString(e));
	}
	
	private void logStackTrace(Exception e){
		log(Level.FINER, Core.stackTraceToString(e));
	}
	
	private void log(String msg){
		this.log(Level.FINER, msg);
	}

}
