package jobExecutors;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.logging.Level;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import core.Core;
import core.Job;
import core.errorHelper.ErrorObject;

/**
 * Testing complete: OK
 * 
 * @author Thomas Rudfoss
 *
 */
public class TagNNoteJobExecutor implements JobExecutorInterface {

	public static final int RETURNCODE_NO_RESULT = 10;
	public static final int RETURNCODE_ELEMENT_EXISTS = 20;
	
	private static String[] jobsICanDoList = new String[]{"createTag", "tagItem", "createNoteForItem", "getNotesForItem", "getTagsForItem"}; 

	public static String[] jobsICanDo(){
		return jobsICanDoList.clone();
	}
	
	private Job theJob = null;
	private JSONObject data = null;
	private JSONObject details = null;
	
	@Override
	public void doJob(Job job) {
		log("Going to work!");
		
		theJob = job;
		try {
			theJob.request = theJob.request.getJSONObject("request");
		} catch (JSONException e1) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_INVALID_REQUEST_FORMAT, theJob);
			log(Level.FINE, "Could not grab request node.");
			logStackTrace(Level.FINE, e1);
			return;
		}
		
		if ( !hasProperRights() ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_ACCESS_DENIED, theJob);
			return;
		}
		
		try {
			theJob.internalResult = new JSONObject("{\"response\":{\"responsespec\":{\"status\":0,\"statusmessage\":\"OK\"}}}");
		} catch (JSONException e) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
			log(Level.SEVERE, "Could not initialize result object!");
			logStackTrace(Level.SEVERE, e);
			return;
		}
		
		try {
			if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[0]) ){
				data = theJob.request.getJSONObject("data");
				this.createTag();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[1]) ){
				details = theJob.request.getJSONObject("details");
				this.tagItem();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[2]) ){
				details = theJob.request.getJSONObject("details");
				data = theJob.request.getJSONObject("data");
				this.createNoteForItem();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[3]) ){
				details = theJob.request.getJSONObject("details");
				this.getNotesForItem();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[4]) ){
				details = theJob.request.getJSONObject("details");
				this.getTagsForItem();
			} else {
				log(Level.SEVERE, "Operation not recognized: "+theJob.getJobTask());
			}
			
			theJob.state = Job.STATE_RESULT_SET;
		} catch ( JSONException e ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_MISSING_DATA, theJob);
			log("Missing data for the request.");
			logStackTrace(e);
		} catch (SQLException e) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
			log(Level.WARNING, "SQL Exception occured when performing operation.");
			logStackTrace(Level.WARNING, e);
		}
		
		log("Finished work!");
	}
	
	private void createTag() throws SQLException, JSONException{
		//Tag already exists?
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT name FROM Tag WHERE name=? LIMIT 1");
		stmt.setString(1, data.getString("name"));
		stmt.execute();
		
		ResultSet rs = stmt.getResultSet();
		if ( rs != null ){
			int count = 0;
			while ( rs.next() ){
				count++;
			}
			
			if ( count > 0 ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "Tag already exists!", TagNNoteJobExecutor.RETURNCODE_ELEMENT_EXISTS);
				log(Level.FINER,"Tag already exists");
				return;
			}
		}
		
		stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO Tag (name, description) VALUES (?, ?)");
		stmt.setString(1, data.getString("name"));
		stmt.setString(2, data.getString("description"));
		stmt.execute();
	}
	
	private void tagItem() throws SQLException, JSONException{
		//Check if the user has tagged the item already with the given tag
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT identifier FROM TagElement WHERE tag_identifier=? AND elementTableName=? AND element_identifier=? AND taggedBy_user_identifier=? LIMIT 1");
		stmt.setInt(1, details.getInt("tagId"));
		stmt.setString(2, details.getString("tableName"));
		stmt.setInt(3, details.getInt("elementId"));
		stmt.setInt(4, theJob.userData.getId());
		stmt.execute();
		
		ResultSet rs = stmt.getResultSet();
		if ( rs != null ){
			int count = 0;
			while ( rs.next() ){
				count++;
			}
			
			if ( count > 0 ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "The person has already tagged the item with the given tag!", TagNNoteJobExecutor.RETURNCODE_ELEMENT_EXISTS);
				log(Level.FINER,"Person has already tagged this item with this tag");
				return;
			}
		}
		
		stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO TagElement (tag_identifier, elementTableName, element_identifier, taggedBy_user_identifier) VALUES (?, ?, ?, ?)");
		stmt.setInt(1, details.getInt("tagId"));
		stmt.setString(2, details.getString("tableName"));
		stmt.setInt(3, details.getInt("elementId"));
		stmt.setInt(4, theJob.userData.getId());
		stmt.execute();
	}
	
	private void createNoteForItem() throws SQLException, JSONException{
		//Create note
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO Note (person_identifier, note) VALUES (?, ?)");
		stmt.setInt(1, details.getInt("creatorPersonId"));
		stmt.setString(2, data.getString("note"));
		stmt.execute();
		
		//Grab last id (this is connection dependent and therefore unaffected by any concurrent processes)
		stmt = theJob.dbConnection.newPreparedStatement("SELECT LAST_INSERT_ID() FROM Note");
		stmt.execute();
		ResultSet rs = stmt.getResultSet();
		int noteId = -1;
		if ( rs == null ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
			log(Level.SEVERE, "Could not grab the last id from Note. Cannot connect the note.");
			return;
		}
		rs.next();
		noteId = rs.getInt(1);
		rs.close();
		
		//Insert elementNote to tag an item
		stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO ElementNote (note_identifier, elementTableName, element_identifier) VALUES (?, ?, ?)");
		stmt.setInt(1, noteId);
		stmt.setString(2, details.getString("tableName"));
		stmt.setInt(3, details.getInt("creatorPersonId"));
		stmt.execute();
	}
	
	private void getNotesForItem() throws SQLException, JSONException{
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT n.note, n.person_identifier FROM Note as n, ElementNote as en WHERE en.elementTableName=? AND en.element_identifier=? AND n.identifier=en.note_identifier");
		stmt.setString(1, details.getString("tableName"));
		stmt.setInt(2, details.getInt("elementId"));
		stmt.execute();
		
		ResultSet rs = stmt.getResultSet();
		JSONArray dataRows = new JSONArray();
		theJob.internalResult.getJSONObject("response").put("data", (new JSONObject()).put("row", dataRows));
		
		if ( rs == null ){
			return;
		}
		
		while ( rs.next() ){
			JSONObject thisObj = new JSONObject();
			
			thisObj.put("note", rs.getString(1));
			thisObj.put("personId", rs.getInt(2));
			
			dataRows.put(thisObj);
		}
	}
	
	private void getTagsForItem() throws SQLException, JSONException{
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT t.name, t.description FROM Tag as t, TagElement as te WHERE te.elementTableName=? AND te.element_identifier=? AND t.identifier=te.tag_identifier");
		stmt.setString(1, details.getString("tableName"));
		stmt.setInt(2, details.getInt("elementId"));
		stmt.execute();
		
		ResultSet rs = stmt.getResultSet();
		JSONArray dataRows = new JSONArray();
		theJob.internalResult.getJSONObject("response").put("data", (new JSONObject()).put("row", dataRows));
		
		if ( rs == null ){
			return;
		}
		
		while ( rs.next() ){
			JSONObject thisObj = new JSONObject();
			
			thisObj.put("note", rs.getString(1));
			
			dataRows.put(thisObj);
		}
	}
	
	private boolean hasProperRights(){
		if ( theJob.userData.isAdmin() ) return true; //Admins always have access

		//"createTag", "tagItem", "createNoteForItem", "getNotesForItem"
		if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[0]) ){
			if ( theJob.userData.hasRole("tagsAdmin") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[1]) ){
			if ( theJob.userData.hasRole("tagsAdmin") ) return true;
			if ( theJob.userData.hasRole("doTag") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[2]) ){
			return true; //Anyone can create notes
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[3]) ){
			return true; //Anyone can display notes
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[4]) ){
			return true; //Anyone can display tags
		}
		
		return false;
	}
	
	private void log(Level level, String msg){
		Core.getCore().getPluginLogger().log(level, "TagNNoteJobExecutor: "+msg);
	}
	
	private void logStackTrace(Level level, Exception e){
		log(level, Core.stackTraceToString(e));
	}
	
	private void logStackTrace(Exception e){
		log(Level.FINER, Core.stackTraceToString(e));
	}
	
	private void log(String msg){
		this.log(Level.FINER, msg);
	}

}
