package jobExecutors;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.logging.Level;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import core.Core;
import core.Job;
import core.errorHelper.ErrorObject;
import core.userData.UserData;
import core.userData.exceptions.DatabaseConnectionException;
import core.userData.exceptions.TokenNotFoundException;
import core.userData.exceptions.UserIdNotFoundException;

public class UserJobExecutor implements JobExecutorInterface {

	public static final int RETURNCODE_NO_RESULT = 10;
	public static final int RETURNCODE_ELEMENT_EXISTS = 20;
	public static final int RETURNCODE_NO_ELEMENT_EXISTS = 30;
	public static final int RETURNCODE_CANNOT_DELETE_SELF = 40;
	public static final int RETURNCODE_NOT_ADMIN = 50;
	
	public static final int RETURNCODE_INVALID_OPTION = 60;
	
	private static String[] jobsICanDoList = new String[]{"createUser", "removeUser", "createAccessRoleInstance", "connectAccessRoleInstanceToUser", "removeAccessRoleInstanceFromUser", "listUsers", "listAccessRoles"};

	public static String[] jobsICanDo(){
		return jobsICanDoList.clone();
	}
	
	private Job theJob = null;
	private JSONObject data = null;
	private JSONObject details = null;
	
	@Override
	public void doJob(Job job) {
		log("Going to work!");
		
		theJob = job;
		try {
			theJob.request = theJob.request.getJSONObject("request");
		} catch (JSONException e1) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_INVALID_REQUEST_FORMAT, theJob);
			log(Level.FINE, "Could not grab request node.");
			logStackTrace(Level.FINE, e1);
			return;
		}
		
		if ( !hasProperRights() ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_ACCESS_DENIED, theJob);
			return;
		}
		
		try {
			theJob.internalResult = new JSONObject("{\"response\":{\"responsespec\":{\"status\":0,\"statusmessage\":\"OK\"}}}");
		} catch (JSONException e) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
			log(Level.SEVERE, "Could not initialize result object!");
			logStackTrace(Level.SEVERE, e);
			return;
		}
		
		try {
			if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[0]) ){
				data = theJob.request.getJSONObject("data");
				this.createUser();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[1]) ){
				details = theJob.request.getJSONObject("details");
				this.removeUser();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[2]) ){
				details = theJob.request.getJSONObject("details");
				this.createAccessRoleInstance();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[3]) ){
				details = theJob.request.getJSONObject("details");
				data = theJob.request.getJSONObject("data");
				this.connectAccessRoleInstanceToUser();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[4]) ){
				details = theJob.request.getJSONObject("details");
				this.removeAccessRoleInstanceFromUser();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[5]) ){
				this.listUsers();
			} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[6]) ){
				this.listAccessRoles();
			}
			
			theJob.state = Job.STATE_RESULT_SET;
		} catch ( JSONException e ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_MISSING_DATA, theJob);
			log("Missing data for the request.");
			logStackTrace(e);
		} catch (SQLException e) {
			ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
			log(Level.WARNING, "SQL Exception occured when inserting into person.");
			logStackTrace(Level.WARNING, e);
		}
		
		log("Finished work!");
	}
	
	private void createUser() throws SQLException, JSONException{
		//First: If person or institute identifier is set check that they exist
		Integer personId = null;
		if ( data.has("person_identifier") && (data.getString("person_identifier").length() > 0) ){
			personId = data.getInt("person_identifier"); 
			PreparedStatement userExists = theJob.dbConnection.newPreparedStatement("SELECT identifier FROM Person WHERE identifier=? LIMIT 1");
			userExists.setInt(1, personId);
			userExists.execute();
			ResultSet rs = userExists.getResultSet();
			if ( rs == null ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No person element with this id exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
				log(Level.FINER,"No person element with id "+personId+" exists.");
				return;
			} else {
				if ( !rs.next() ){
					ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No person element with this id exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
					log(Level.FINER,"No person element with id "+personId+" exists.");
					return;
				}
			}
		}
		
		Integer instituteId = null;
		if ( data.has("institute_identifier") && (data.getString("institute_identifier").length() > 0) ){
			instituteId = data.getInt("institute_identifier"); 
			PreparedStatement instituteExists = theJob.dbConnection.newPreparedStatement("SELECT identifier FROM Institute WHERE identifier=? LIMIT 1");
			instituteExists.setInt(1, instituteId);
			instituteExists.execute();
			ResultSet rs = instituteExists.getResultSet();
			if ( rs == null ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No institute element with this id exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
				log(Level.FINER,"No institute element with id "+instituteId+" exists.");
				return;
			} else {
				if ( !rs.next() ){
					ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No institute element with this id exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
					log(Level.FINER,"No institute element with id "+instituteId+" exists.");
					return;
				}
			}
		}

		//Check that a user with this token does not already exist
		PreparedStatement tokenExists = theJob.dbConnection.newPreparedStatement("SELECT identifier FROM User WHERE token=? LIMIT 1");
		tokenExists.setString(1, data.getString("token"));
		tokenExists.execute();
		ResultSet rs = tokenExists.getResultSet();
		if ( rs != null ){			
			if ( rs.next() ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "User with this token already exists", UserJobExecutor.RETURNCODE_ELEMENT_EXISTS);
				log(Level.FINER,"User with this token already exists");
				return;
			}
		}
		
		//Finally ready to actually create the user :P
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO User (token, person_identifier, institute_identifier) VALUES(?, ?, ?)");
		stmt.setString(1, data.getString("token"));
		if ( personId != null ) stmt.setInt(2, personId); else stmt.setNull(2, Types.NULL);
		if ( instituteId != null ) stmt.setInt(3, instituteId); else stmt.setNull(3, Types.NULL);
		stmt.execute();
			
		//Return the new users id
		theJob.internalResult.getJSONObject("response").put("data", (new JSONObject()).put("userid", this.getLastInsertId("User")) );
	}
	
	private void removeUser() throws SQLException, JSONException{
		//Grab user info
		UserData ud = null;
		if ( details.has("userid") ){
			try {
				ud = UserData.constructFromId(details.getInt("userid"));
			} catch (DatabaseConnectionException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
				log(Level.WARNING,"Internal server error occured.");
				logStackTrace(Level.WARNING, e);
				return;
			} catch (UserIdNotFoundException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No user with this ID exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
				log(Level.FINER, "No user with given id exists. Cannot delete.");
				return;
			}
		} else if ( details.has("token") ){
			try {
				ud = UserData.constructFromToken(details.getString("token"));
			} catch (DatabaseConnectionException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
				log(Level.WARNING,"Internal server error occured.");
				logStackTrace(Level.WARNING, e);
				return;
			} catch (TokenNotFoundException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No user with this token exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
				log(Level.FINER, "No user with given token exists. Cannot delete.");
				return;
			}
		}
		
		if ( ud.getId() == theJob.userData.getId() ){
			//You cannot remove yourself
			ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "You cannot delete yourself", UserJobExecutor.RETURNCODE_CANNOT_DELETE_SELF);
			log(Level.FINER, "Cannot delete yourself");
			return;
		}
		
		if ( ud.isAdmin() && !theJob.userData.isAdmin() ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "You need to be admin!", UserJobExecutor.RETURNCODE_NOT_ADMIN);
			log(Level.FINER, "Admin delete of not admin fail!");
			return;
		}

		if ( ud == null ){
			throw new JSONException("Missing data!");
		}
		
		//Begin cleaning up
		//ResourceRight
		PreparedStatement resourceRights = theJob.dbConnection.newPreparedStatement("DELETE FROM ResourceRight WHERE user_identifier=?");
		resourceRights.setInt(1, ud.getId());
		resourceRights.execute();
		
		//UserRoleInstance
		PreparedStatement userRoleInstances = theJob.dbConnection.newPreparedStatement("DELETE FROM UserRoleInstance WHERE user_identifier=?");
		userRoleInstances.setInt(1, ud.getId());
		userRoleInstances.execute();
		
		//And finally delete the user
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("DELETE FROM User WHERE identifier=?");
		stmt.setInt(1, ud.getId());
		stmt.execute();
	}
	
	private void createAccessRoleInstance() throws SQLException, JSONException{		
		//First check if the exact instance already exists
		//Grab all instance entries
		PreparedStatement arIn = theJob.dbConnection.newPreparedStatement("SELECT identifier FROM RoleInstance WHERE accessRole_identifier=?");
		arIn.setInt(1, details.getInt("roleid"));
		arIn.execute();
		
		//Grab all optionNames for a role
		PreparedStatement opData = theJob.dbConnection.newPreparedStatement("SELECT identifier, optionName FROM RoleOption WHERE accessRole_identifier=?");
		opData.setInt(1, details.getInt("roleid"));
		opData.execute();
		
		//Grab all option names for the access role (options that MUST be filled)
		HashMap<String, DataContainer> optionData = new HashMap<String, DataContainer>();
		while ( opData.getResultSet().next() ){
			DataContainer c = new DataContainer();
			c.d.put("opid", opData.getResultSet().getInt(1));
			c.d.put("name", opData.getResultSet().getString(2));
			
			optionData.put((String) c.d.get("name"), c);
		}
		
		//Grab all option values from request

		try {
			data = theJob.request.getJSONObject("data");
		} catch ( JSONException e ){
			//Apparantly data is not a JSONObject
			data = new JSONObject();
		}
		if ( data.has("option") && data.getJSONArray("option").length() > 0 ){
			//Grab all option values from the request and identify by name
			for ( int i=0; i<data.getJSONArray("option").length(); i++ ){
				JSONObject thisObj = data.getJSONArray("option").getJSONObject(i);
				if ( !optionData.containsKey(thisObj.getString("name")) ){
					ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "The option "+thisObj.getString("name")+" is not defined for the AccessRole with id "+details.getInt("roleid"), UserJobExecutor.RETURNCODE_INVALID_OPTION);
					log(Level.FINER,"Option "+thisObj.getString("name")+" not defined for AccessRole with id "+details.getInt("roleid"));
					return;
				}
				DataContainer c = optionData.get(thisObj.getString("name"));
				c.d.put("reqval", thisObj.getString("value"));
			}
		}
		
		//Do a quick check to see that all options are set
		Iterator<String> it = optionData.keySet().iterator();
		while ( it.hasNext() ){
			DataContainer c = optionData.get(it.next()); 
			if ( !( c.d.containsKey("reqval") ) ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "The option "+c.getStr("name")+" is not present in request", UserJobExecutor.RETURNCODE_INVALID_OPTION);
				log(Level.FINER, "Option missing from request.");
				return;
			}
		}

		while ( arIn.getResultSet().next() ){
			//At least one instance exists. Need to check it
			int instanceId = arIn.getResultSet().getInt(1);			
			
			//Compare all option values for this instance to the ones given in the query
			PreparedStatement optVals = theJob.dbConnection.newPreparedStatement("SELECT rov.optionValue " +
					"FROM RoleOptionValue AS rov " +
					"WHERE rov.roleOption_identifier=? AND rov.roleInstance_identifier=? LIMIT 1");
			it = optionData.keySet().iterator();
			int equalCount = 0;
			while ( it.hasNext() ){
				DataContainer c = optionData.get(it.next());
				optVals.setInt(1, c.getInt("opid"));
				optVals.setInt(2, instanceId);
				optVals.execute();
				optVals.getResultSet().next();
				if ( c.getStr("reqval").equals(optVals.getResultSet().getString(1)) ){
					//This entry already exists
					equalCount++;
				}
			}
			
			if ( optionData.size() == equalCount ){
				//This instance already exists
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "An instance of this role with the exact same option values exists.", UserJobExecutor.RETURNCODE_ELEMENT_EXISTS);
				log(Level.FINER, "Attempt to create dupliate AccessRole instance failed.");
				return;
			}
		}
		
		//Finally we can actually create the instance
		//Here we know that all options are present and that the instance does not already exist
		PreparedStatement instStmt = theJob.dbConnection.newPreparedStatement("INSERT INTO RoleInstance (accessRole_identifier) VALUES (?)");
		PreparedStatement opStmt = theJob.dbConnection.newPreparedStatement("INSERT INTO RoleOptionValue (roleOption_identifier, roleInstance_identifier, optionValue) VALUES (?, ?, ?)");
		instStmt.setInt(1, details.getInt("roleid"));
		instStmt.execute();
		int instId = this.getLastInsertId("RoleInstance");
		
		it = optionData.keySet().iterator();
		while ( it.hasNext() ){
			DataContainer c = optionData.get(it.next());
			opStmt.setInt(1, c.getInt("opid"));
			opStmt.setInt(2, instId);
			opStmt.setString(3, c.getStr("reqval"));
			opStmt.execute();
		}
		
		theJob.internalResult.getJSONObject("response").put("data", (new JSONObject()).put("instanceid", instId));
		//AND DONE!
	}
	
	/**
	 * Used for simple data storage.
	 * 
	 * @author Thomas Rudfoss
	 *
	 */
	class DataContainer{
		
		public HashMap<String, Object> d;
		
		public DataContainer(){
			d = new HashMap<String, Object>();
		}
		
		public int getInt(String key){
			return (Integer) d.get(key);
		}
		
		public String getStr(String key){
			return (String) d.get(key);
		}
	}
	
	private void connectAccessRoleInstanceToUser() throws SQLException, JSONException{
		//TODO Ensure only admins can connect admin roles
		
		//Grab user info
		UserData ud = null;
		if ( details.has("userid") ){
			try {
				ud = UserData.constructFromId(details.getInt("userid"));
			} catch (DatabaseConnectionException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
				log(Level.WARNING,"Internal server error occured.");
				logStackTrace(Level.WARNING, e);
				return;
			} catch (UserIdNotFoundException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No user with this ID exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
				log(Level.FINER, "No user with given id exists. Cannot delete.");
				return;
			}
		} else if ( details.has("token") ){
			try {
				ud = UserData.constructFromToken(details.getString("token"));
			} catch (DatabaseConnectionException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
				log(Level.WARNING,"Internal server error occured.");
				logStackTrace(Level.WARNING, e);
				return;
			} catch (TokenNotFoundException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No user with this token exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
				log(Level.FINER, "No user with given token exists. Cannot delete.");
				return;
			}
		}
		
		//Grab access role to check that it exists and if it is the admin (root)
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT ar.name FROM RoleInstance AS ri, AccessRole AS ar WHERE ri.identifier=? AND ar.identifier=ri.accessRole_identifier LIMIT 1");
		stmt.setInt(1, data.getInt("accessRoleInstanceId"));
		stmt.execute();
		if ( !stmt.getResultSet().next() ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "The AccessRole or AccessRole instance does not exist", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
			log(Level.FINER, "Attempted to set instance for role that does not exists");
			return;
		} else {
			if ( !theJob.userData.isAdmin() && (stmt.getResultSet().getString(1).equalsIgnoreCase("root")) ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "You need to be admin!", UserJobExecutor.RETURNCODE_NOT_ADMIN);
				log(Level.FINER, "Assign admin failed");
				return;
			}
		}
		
		//Set the instance
		stmt = theJob.dbConnection.newPreparedStatement("INSERT INTO UserRoleInstance (user_identifier, roleInstance_identifier, assignedDateTime, expireDateTime) VALUES (?, ?, ?, ?)");
		stmt.setInt(1, ud.getId());
		stmt.setInt(2, data.getInt("accessRoleInstanceId"));
		stmt.setString(3, getSQL(new Date()));
		stmt.setNull(4, Types.NULL);
		
		if ( data.has("expireTime") && (data.getString("expireTime").length() == 19) ){
			stmt.setString(4, data.getString("expireTime"));
		}
		
		try {
			stmt.execute();
		} catch ( SQLException e ){
			//Should only happen if the expireTime format is invalid
			ErrorObject.getErrorForCode(ErrorObject.CODE_INVALID_REQUEST_FORMAT, theJob, "The expireTime format is invalid");
			log(Level.FINER, "ExpireTime format is invalid. See Finest log level for exception trace");
			logStackTrace(Level.FINEST, e);
		}
	}
	
	private void removeAccessRoleInstanceFromUser() throws SQLException, JSONException{
		UserData ud = null;
		if ( details.has("userid") ){
			try {
				ud = UserData.constructFromId(details.getInt("userid"));
			} catch (DatabaseConnectionException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
				log(Level.WARNING,"Internal server error occured.");
				logStackTrace(Level.WARNING, e);
				return;
			} catch (UserIdNotFoundException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No user with this ID exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
				log(Level.FINER, "No user with given id exists. Cannot delete.");
				return;
			}
		} else if ( details.has("token") ){
			try {
				ud = UserData.constructFromToken(details.getString("token"));
			} catch (DatabaseConnectionException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_SERVER_ERROR, theJob);
				log(Level.WARNING,"Internal server error occured.");
				logStackTrace(Level.WARNING, e);
				return;
			} catch (TokenNotFoundException e) {
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "No user with this token exists", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
				log(Level.FINER, "No user with given token exists. Cannot delete.");
				return;
			}
		}
		
		//Check if the user actually has that instance connected to it.
		PreparedStatement userRI = theJob.dbConnection.newPreparedStatement("SELECT * FROM UserRoleInstance WHERE user_identifier=? AND roleInstance_identifier=? LIMIT 1");
		userRI.setInt(1, ud.getId());
		userRI.setInt(2, details.getInt("accessRoleInstanceId"));
		userRI.execute();
		if ( !userRI.getResultSet().next() ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "User does not have the accessroleinstance id connected to it.", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
			log(Level.FINER, "Attempted to remove instance that is not connected to user");
			return;
		}

		//Grab access role to check that it exists and if it is the admin (root)
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT ar.name FROM RoleInstance AS ri, AccessRole AS ar WHERE ri.identifier=? AND ar.identifier=ri.accessRole_identifier LIMIT 1");
		stmt.setInt(1, details.getInt("accessRoleInstanceId"));
		stmt.execute();
		if ( !stmt.getResultSet().next() ){
			ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "The AccessRole or AccessRole instance does not exist", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
			log(Level.FINER, "Attempted to remove instance for role that does not exists");
			return;
		} else {
			if ( !theJob.userData.isAdmin() && (stmt.getResultSet().getString(1).equalsIgnoreCase("root")) ){
				ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "You need to be admin!", UserJobExecutor.RETURNCODE_NOT_ADMIN);
				log(Level.FINER, "Attempt to remove admin by non-admin failed.");
				return;
			} else if ( stmt.getResultSet().getString(1).equalsIgnoreCase("root") ){
				//Check that at least one more person has this role
				PreparedStatement admList = theJob.dbConnection.newPreparedStatement("SELECT COUNT(uri.identifier) FROM UserRoleInstance AS uri, RoleInstance AS ri, AccessRole AS ar WHERE " +
						"ar.name=\"root\" AND ri.accessRole_identifier=ar.identifier AND uri.roleInstance_identifier=ri.identifier AND uri.expireDateTime>NOW()");
				admList.execute();
				if ( !admList.getResultSet().next() || (admList.getResultSet().getInt(1) <= 1) ){
					ErrorObject.getErrorForCode(ErrorObject.CODE_UNKNOWN_ERROR, theJob, "The system must have at least one admin", UserJobExecutor.RETURNCODE_NO_ELEMENT_EXISTS);
					log(Level.FINER, "Attempt to remove last admin.");
					return;
				}
			}
		}
		
		//It is now safe to remove the instance
		stmt = theJob.dbConnection.newPreparedStatement("DELETE FROM UserRoleInstance WHERE user_identifier=? AND roleInstance_identifier=?");
		stmt.setInt(1, ud.getId());
		stmt.setInt(2, details.getInt("accessRoleInstanceId"));
		stmt.execute();
	}
	
	private void listUsers() throws SQLException, JSONException{
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT * FROM User");
		stmt.execute();
		
		JSONArray rows = new JSONArray();
		while ( stmt.getResultSet().next() ){
			JSONObject row = new JSONObject();
			for ( int i=1; i<=stmt.getMetaData().getColumnCount(); i++ ){
				row.put(stmt.getMetaData().getColumnLabel(i), stmt.getResultSet().getObject(i));
			}
			
			rows.put(row);
		}
		
		theJob.internalResult.getJSONObject("response").put("data", (new JSONObject()).put("row", rows));
	}
	
	private void listAccessRoles() throws SQLException, JSONException{
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT * FROM AccessRole");
		stmt.execute();
		
		JSONArray rows = new JSONArray();
		while ( stmt.getResultSet().next() ){
			JSONObject row = new JSONObject();
			for ( int i=1; i<=stmt.getMetaData().getColumnCount(); i++ ){
				row.put(stmt.getMetaData().getColumnLabel(i), stmt.getResultSet().getObject(i));
			}
			
			rows.put(row);
		}
		
		theJob.internalResult.getJSONObject("response").put("data", (new JSONObject()).put("row", rows));
	}
	
	private String getSQL(Date d){
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
		
		return sdf.format(d);
	}
	
	/**
	 * Returns the last inserted id for a table
	 * 
	 * @param table The table to check
	 * @return
	 * @throws SQLException
	 */
	private int getLastInsertId(String table) throws SQLException{
		PreparedStatement stmt = theJob.dbConnection.newPreparedStatement("SELECT LAST_INSERT_ID() FROM "+table+" LIMIT 1");
		stmt.execute();

		ResultSet rs = stmt.getResultSet();
		rs.next();
		
		return rs.getInt(1);
	}
	
	private boolean hasProperRights(){
		if ( theJob.userData.isAdmin() ) return true; //Admins always have access
		
		if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[0]) ){
			if ( theJob.userData.hasRole("userAdmin") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[1]) ){
			if ( theJob.userData.hasRole("userAdmin") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[2]) ){
			if ( theJob.userData.hasRole("userAdmin") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[3]) ){
			if ( theJob.userData.hasRole("userAdmin") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[4]) ){
			if ( theJob.userData.hasRole("userAdmin") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[5]) ){
			if ( theJob.userData.hasRole("userAdmin") ) return true;
			if ( theJob.userData.hasRole("viewAll") ) return true;
		} else if ( theJob.getJobTask().equalsIgnoreCase(jobsICanDoList[6]) ){
			if ( theJob.userData.hasRole("userAdmin") ) return true;
			if ( theJob.userData.hasRole("viewAll") ) return true;
		}
		
		return false;
	}
	
	private void log(Level level, String msg){
		Core.getCore().getPluginLogger().log(level, "UserJobExecutor: "+msg);
	}
	
	private void logStackTrace(Level level, Exception e){
		log(level, Core.stackTraceToString(e));
	}
	
	private void logStackTrace(Exception e){
		log(Level.FINER, Core.stackTraceToString(e));
	}
	
	private void log(String msg){
		this.log(Level.FINER, msg);
	}

}
