package resultTranslators;

import java.util.Iterator;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import core.Core;
import core.Job;

/**
 * Translates into comma separated values if possible.
 * If the data has more levels, an error message is returned
 * The root node is always ignored.
 * 
 * Example:
 * <root>
 * 		<elem>A</elem>
 * 		<elem>B</elem>
 * 		<elem>C</elem>
 * </root>
 * Equals:
 * elem,elem,elem
 * A,B,C
 * 
 * <root>
 * 		<aRow>
 * 			<id>429374</id>
 * 			<firstname>Ola</firstname>
 * 			<lastname>Nordmann</lastname>
 *		</aRow>
 *		<aRow>
 *			<id>827592</id>
 * 			<firstname>Kari</firstname>
 * 			<lastname>Nordkvinne</lastname>
 *		</aRow>
 * </root>
 * Equals:
 * errormessage
 * "multileveled result, cannot parse"
 * 
 * 
 * @author Thomas Rudfoss
 *
 */
public class CSVTranslator implements ResultTranslatorInterface {

	public static String[] iCanTranslateTo(){
		return new String[]{"csv"};
	}
	
	private final String ERROR_NO_RESULT = "errormessage\n\"no result in response\"";
	private final String ERROR_MULTILEVEL_RESULT = "errormessage\n\"multileveled result, cannot parse\"";
	private final String ERROR_NO_DATA = "errormessage\n\"no data in result\"";
	private final String ERROR_INTERNAL = "errormessage\n\"internal translator error\"";
	
	@Override
	public void translateJob(Job job, String toFormat) {
		if ( !job.internalResult.has("response") ){
			job.result = ERROR_NO_RESULT;
			return;
		}
		
		JSONObject resp = null;
		try {
			 resp = job.internalResult.getJSONObject("response");
		} catch (JSONException e) {
			Core.getCore().getPluginLogger().warning("CSVTranslator: Response element either missing or not a JSONObject!");
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
			job.result = ERROR_INTERNAL;
			return;
		}
		
		if ( !resp.has("data") ){
			job.result = ERROR_NO_DATA;
			return;
		}
		
		JSONArray data = null;
		try {
			 data = resp.getJSONArray("response");
		} catch (JSONException e) {
			Core.getCore().getPluginLogger().warning("CSVTranslator: Data element is either missing or not a JSONArray!");
			Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
			job.result = ERROR_INTERNAL;
			return;
		}
		
		if ( data.length() < 1 ){
			job.result = ERROR_NO_DATA;
			return;
		}
		
		StringBuilder csvHeadings = new StringBuilder();
		StringBuilder csvData = new StringBuilder();
		//Parse the data
		for ( int i=0; i<data.length(); i++ ){
			JSONObject row = null;
			try {
				 row = data.getJSONObject(i);
			} catch (JSONException e) {
				Core.getCore().getPluginLogger().warning("CSVTranslator: Error extracting row "+i+" from data!");
				Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
				job.result = ERROR_INTERNAL;
				return;
			}
			
			//Process a single row
			Iterator<?> it = row.keys();
			while ( it.hasNext() ){				
				String key = (String)it.next();
				String val = "";
				try {
					val = row.getString(key);
				} catch ( JSONException e ){
					Core.getCore().getPluginLogger().warning("CSVTranslator: Cannot translate data. One or more elements are not leaf nodes");
					Core.getCore().getPluginLogger().warning(Core.stackTraceToString(e));
					job.result = ERROR_MULTILEVEL_RESULT;
					return;
				}
				
				if ( val.contains("\"") ) val = val.replaceAll("\"", "\"\"");
				csvData.append("\""+val+"\",");
				if ( i == 0 ){
					csvHeadings.append("\""+key+"\",");
				}
			}
			
			//Remove unnecessary comma from end
			if ( i == 0 ) csvHeadings.deleteCharAt(csvHeadings.length()-1);
			csvData.deleteCharAt(csvData.length()-1);
			csvData.append("\n");
		}
		
		job.result = csvHeadings.append("\n").append(csvData).toString();
	}
}
